#------------------------------------------------------------------
# $Copyright: Copyright (c) 2010 Symantec Corporation.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# SYMANTEC CORPORATION.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SYMANTEC CORPORATION.
#
# The Licensed Software and Documentation are deemed to be commercial
# computer software as defined in FAR 12.212 and subject to restricted
# rights as defined in FAR Section 52.227-19 "Commercial Computer
# Software - Restricted Rights" and DFARS 227.7202, "Rights in
# Commercial Computer Software or Commercial Computer Software
# Documentation", as applicable, and any successor regulations. Any use,
# modification, reproduction release, performance, display or disclosure
# of the Licensed Software and Documentation by the U.S. Government
# shall be solely in accordance with the terms of this Agreement.  $
#
# $Header: /project/isv-agent/VProAgents/Informix/Informix.pm,v 1.22 2010/09/13 12:51:06 ssrivast Exp $
#
# $Author: ssrivast $
# $Date: 2010/09/13 12:51:06 $
# $Revision: 1.22 $
#
# Provides:                            Informix
# Author:                              Anshuman Atre
# crAgent Build:                       1.85 2008/08/19 06:59:10
# crAgent Run Date:                    23 Sep, 2008
# Modification History:
# Known bugs/caveats:
#------------------------------------------------------------------

#------------
# Be strict..
#------------
use strict;

#----------------------
# Declare the package..
#----------------------
package Informix;

#-----------------------------------------
# Define module-vendor specific commands..
#-----------------------------------------
use ag_i18n_inc;

#-------------------------------------
# Include required framework modules..
#-------------------------------------
use Sys;            # Support generalized UNIX commands
use Proc;           # For calls to the process routines
use Comms;          # For calls to socket(3C) routines
use Version;        # Version display generator
use VCSagentFW;     # Primary Agent Interface to the ACC Library
use VCSDefines;     # This contains constants that VCS uses

my ( $CLUS_HOME, $sPrefix, $bEnv, $bOnlineNode, @liRunningProcs ) = ();

#------------------------
# Setup VCS/VCSone home..
#------------------------
if ( $ENV { CLUSTER_HOME } ) { # VCS 5.x/VCSOne..
  $CLUS_HOME ||= $ENV { CLUSTER_HOME } ||= "/opt/VRTSvcsone";
}
else { # VCS 4.x..
  $CLUS_HOME ||= $ENV { VCS_HOME } ||= "/opt/VRTSvcs";
}

( $sPrefix ) = $CLUS_HOME =~ /.*(vad|vcs(|one))$/;

my $sCmdHares  = $HARES  ||= "$CLUS_HOME/bin/hares";
my $sCmdHagrp  = $HAGRP  ||= "$CLUS_HOME/bin/hagrp";
my $sCmdHatype = $HATYPE ||= "$CLUS_HOME/bin/hatype";
my $sCmdHasys  = $HASYS  ||= "$CLUS_HOME/bin/hasys";

#--------------------------------
# Include standard PERL modules..
#--------------------------------
use Fcntl;
use File::Path qw ( mkpath );
use POSIX qw ( uname );
use Socket;

#-----------------------------------------------------------------------
# Add our version number to the version hash and check if versions are
# requested..
#-----------------------------------------------------------------------
my $sVer = "\$Id: Informix.pm,v 1.22 2010/09/13 12:51:06 ssrivast Exp $::ETVHeader";
Version::TellVersion ( $sVer );

#--------------------------------
# Define Informix subroutines..
#--------------------------------
use constant O_PIDFILE => 1 << 0;              # PID file..
use constant O_IPCFILE => 1 << 1;              # SHM file..
use constant O_HISTORY => O_PIDFILE|O_IPCFILE; # Both..

=pod

 -------------------------------------------------------------------------
 Function :  bool ArgsValid ( IN String EntryPointName,
                              IN String EnvFile,
                              IN String InformixDir,
                              IN String DBServerName,
                              IN String ServerType,
                              IN String SDSInstances,
                              IN Integer SecondLevelMonitor,
                              IN String MonitorProgram );

 Description : Validates the ArgList passed to the entry point.

 Parameters :

 $sEntryPointName               Entry pt name
 $localEnvFile                  Absolute path to environment file to source
 $localInformixDir              Absolute path to installation directory
 $localDBServerName             Unique name of this instance of the database server
 $localServerType 		Standard/HDR/SharedDisk
 $localSDSInstances 		List of Connected SDS
 $localSecondLevelMonitor       Cycle count to run in-depth monitoring
 $localMonitorProgram           Full path to external monitor program
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ArgsValid {
  #----------------------------------------
  # Set up local vars from passed in ones..
  #----------------------------------------
  my ( $sEntryPointName,
       $localEnvFile,
       $localInformixDir,
       $localDBServerName,
       $localServerType,
       $localSDSInstances,
       $localSecondLevelMonitor,
       $localMonitorProgram ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20001, "Subroutine <%s> called with:
  EntryPointName           [%s]
  EnvFile                  [%s]
  InformixDir              [%s]
  DBServerName             [%s]
  ServerType 	           [%s]
  SecondLevelMonitor       [%s]
  MonitorProgram           [%s]",

  $sMyName,  $sEntryPointName,  $localEnvFile,   $localInformixDir,
  $localDBServerName, $localServerType, $localSecondLevelMonitor,
  $localMonitorProgram );

  #-------------------------------------
  # Set return code to hopeful success..
  #-------------------------------------
  my $iReturnCode = 1;

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  my %hsZones = map { $_ => 1 } qw ( 5.10 5.11 ); # For Solaris zones..
  my ( @lsMissing, %hsArgList ) = ();
  my @lsArgs =
  qw ( EntryPointName       EnvFile       InformixDir       DBServerName
       SecondLevelMonitor MonitorProgram );

  @hsArgList { @lsArgs } = @_; # Assign..

  #----------------------------------------
  # Reqd. attributes based on entry point..
  #----------------------------------------
  my @lsOnlineArgs  = qw ( DBServerName EnvFile InformixDir );
  my @lsOfflineArgs = qw ( DBServerName EnvFile InformixDir );
  my @lsMonitorArgs = qw ( DBServerName InformixDir );
  my @lsCleanArgs   = qw ( DBServerName );

  my %hlsNeeded = ( 'online'     => \@lsOnlineArgs,
                    'offline'    => \@lsOfflineArgs,
                    'monitor'    => \@lsMonitorArgs,
                    'clean'      => \@lsCleanArgs,
                    'switchprim' => \@lsOnlineArgs,
                    'isprimary'  => \@lsOnlineArgs );

  my @lsReqdArgs = @{ $hlsNeeded { $sEntryPointName } };

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20002,
    "<%s> The following mandatory fields were not set [%s], Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return, if failed..

  #-----------------------------
  # Remove SLM file, if exists..
  #-----------------------------
  if ( $sEntryPointName ne 'monitor' ) {
    VCSagentFW::RunSecondLevelMonitor ( 1 ); # Delete..
  }

  #-------------------------------
  # Individual validation starts..
  #-------------------------------
  
  #----------------------------------
  # Check if user "informix" exists..
  #----------------------------------
  my $sUser = "informix";
  if ( ! defined getpwnam $sUser ) {
    &logError ( 20003,
    "<%s> [%s] not a valid UNIX login", $sMyName, $sUser );
    $iReturnCode = 0; # Failure..
  }

  #---------------------------------------
  # Validation for online entry point(s)..
  #---------------------------------------
  if ( $sEntryPointName eq "online" ) {
    #-----------------------------
    # Validate parameter EnvFile..
    #-----------------------------
    if ( ! -e $localEnvFile ) {
      &logError ( 20004, "<%s> EnvFile [%s] file does not exist",
      $sMyName, $localEnvFile );
      $iReturnCode = 0; # Failure..
    }

    #---------------------------------
    # Validate parameter InformixDir..
    #---------------------------------
    if ( ! -d $localInformixDir ) {
      &logError ( 20005, "<%s> InformixDir [%s] directory does not exist",
      $sMyName, $localInformixDir );
      $iReturnCode = 0; # Failure..
    }

    #------------------------------------
    # Validate parameter MonitorProgram..
    #------------------------------------
    if ( $localMonitorProgram =~ /(\S+)/ && # Defined..
         ! -x $1 ) {
      &logError ( 20006, "<%s> MonitorProgram [%s] is not executable",
      $sMyName, $1 );
      $iReturnCode = 0; # Failure..
    }
  }

  #-----------------------------------------------
  # Validate "onstat" on online node for HP-UX and
  # Solaris..
  #-----------------------------------------------
  my $sCmd = "$localInformixDir/bin/onstat";
  if ( $^O eq 'solaris' || $^O eq 'hpux' ) {
    $bEnv = 0; # Set global..

    &logTrace ( 20263,
    "<%s> Running inside OS [%s]. Env flag set to [%s]",
    $sMyName, $^O, $bEnv );
  }
  else {  
    ++ $bEnv;
  }

  if ( ! $bEnv &&
       VCSagentFW::AmIFaulting ( VCSDef_MonitorExit::OFFLINE ) ) {
    if ( ! -e $sCmd ) {
      &logError ( 20007, "<%s> Informix monitor utility [%s] does not exist",
      $sMyName, $sCmd );
      $iReturnCode = 0; # Failure..
    }
    else {
      ++ $bOnlineNode; # Onstat exists on ONLINE node..
    }
  }

  #---------------------------------------------
  # Validate ServerType and SDSInstances
  #---------------------------------------------

  if ( $localServerType eq "SharedDisk" &&
       0 == scalar @$localSDSInstances ) {
    &logError ( 20243,
    "<%s> The SDSInstances list must not be empty if ServerType is ShareDisk",
    $sMyName);
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode ||
  &logWarn ( 20008,
  "<%s> Some of the attributes could not be validated. Returning [%s]",
  $sMyName, $iReturnCode );

  return $iReturnCode ;
}



=pod

 -------------------------------------------------------------------------
 Function : bool Start ( IN String DBServerName,
                         IN String EnvFile,
                         IN String InformixDir,
                         IN Integer SecondLevelMonitor );

 Description : Start logic needed for online entry pt

 Parameters :

 $sDBServerName             Unique name of this instance of the database server
 $sEnvFile                  Absolute path to environment file to source
 $sInformixDir              Absolute path to installation directory
 $iSecondLevelMonitor       Cycle count to run in-depth monitoring
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Start {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sDBServerName,
       $sEnvFile,
       $sInformixDir,
       $iSecondLevelMonitor ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20009, "Subroutine <%s> called with:
  DBServerName             [%s]
  EnvFile                  [%s]
  InformixDir              [%s]
  SecondLevelMonitor       [%s]",

  $sMyName,     $sDBServerName,      $sEnvFile,      $sInformixDir,
  $iSecondLevelMonitor );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iSleepTime, $sCommand, $sCommandArgs, $iTillWhen, $iExitCode, $iTimeout,
       @lsCmdout, @liPids ) = ();

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  $iTillWhen = $^T + Informix::GetEngineTimeout ( 1 );
  
  #----------------------------------------------------------------------
  # The utility 'onmode' is known to fail, if it cannot attach to certain
  # shared memory segments. Hence it's failure could occur either because
  # persistent SHMids do not exist, OR the IDs that store the required
  # information were absent and hence the execution failed.
  #
  # In that case we use the IPC history file if it exists, to attempt to
  # clean IPC resources. Atre.. 08 Nov 2008
  #----------------------------------------------------------------------

  #----------------------------------
  # Remove persistent IPC resources..
  #----------------------------------
  $sCommand = "$sInformixDir/bin/onmode";
  $sCommandArgs = "-ky";
  $iSleepTime = 5; # Secs..

  ( @lsCmdout, $iExitCode, $iTimeout ) = ();

  &logTrace ( 20010,
  "<%s> Removing stale IPC resources for Informix instance [%s] with Cmd [%s] Args [%s]",
  $sMyName, $sDBServerName, $sCommand, $sCommandArgs );

  #------------------
  # Request startup..
  #------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20011,
    "<%s> Insufficient time to continue boot up. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = 0; # Failure..
  }
  elsif ( $iReturnCode =
    Informix::ExecLocalCmd ( "informix", 
                             $sEnvFile, 
                             $sCommand, 
                             $sCommandArgs, 
                             \$iExitCode, 
                             $iTimeout, 
                             \@lsCmdout ) ) {
    &logTrace ( 20012,
    "<%s> Cmd [%s] with args [%s] executed successfully. Pausing for [%s] sec.",
    $sMyName, $sCommand, $sCommandArgs, $iSleepTime );

    sleep $iSleepTime; # Pause start-up..
  }

  Informix::RemoveIPCs (); # Ignore exit code..
 
  #---------------------------
  # Re-direct command output..
  #---------------------------
  my $sRdStr = ( getpwnam "informix" ) [-1] =~ /csh$/o ?">& /dev/null":
                                                       "1> /dev/null 2>&1";
  #------------------------------------------
  # Define start command and args variables..
  #------------------------------------------
  $sCommand = "$sInformixDir/bin/oninit";
  $sCommandArgs = "-y $sRdStr";
  ( @lsCmdout, $iExitCode, $iTimeout ) = ();

  &logTrace ( 20013,
  "<%s> Starting Informix instance [%s] with Cmd [%s] Args [%s]",
  $sMyName, $sDBServerName, $sCommand, $sCommandArgs );

  #------------------
  # Request startup..
  #------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20014,
    "<%s> Insufficient time to continue boot up. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = 0; # Failure..
  }
  elsif ( $iReturnCode =
    Informix::ExecLocalCmd ( "informix", 
                             $sEnvFile, 
                             $sCommand, 
                             $sCommandArgs, 
                             \$iExitCode, 
                             $iTimeout, 
                             \@lsCmdout ) ) {
    &logTrace ( 20015,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }

  #----------------------------------------------------------------------
  # Report if SD Primary is not ONLINE but trying to ONLINE SD Secondary.
  #----------------------------------------------------------------------
  if ( $lsCmdout [0] && $lsCmdout [0] =~
     /Can not initialize SDS - SDS Initial Connection Failed/ && 1 == $iExitCode ) {

    &logWarn ( 20183,
    "<%s> SD Primary is NOT online. Please Start the SD Primary first.\n",
    $sMyName );

  }

  #-------------------
  # Return if failed..
  #-------------------
  $iReturnCode || return $iReturnCode;

  #--------------------
  # Delay entry point..
  #--------------------
  $iSleepTime = 10; # Secs..
  
  while ( ( $iTimeout = $iTillWhen - time ) > 10 ) {
    @liPids = (); # Initialise..

    sleep $iSleepTime; # Sec..

    &logTrace ( 20016,
    "<%s> Confirming start-up with [%s] sec in hand",
    $sMyName, $iTimeout - $iSleepTime );

    #---------
    # Do FLM..
    #---------
    if ( VCSDef_MonitorExit::ONLINE ==
         Informix::GetProcState ( \@liPids,
                                  $sDBServerName,
                                  $sEnvFile,
                                  $sInformixDir,
                                  1 ) && (           # Running FFLM..
         $iSecondLevelMonitor  ?                     # If enabled..
         VCSDef_MonitorExit::ONLINE ==               # Online after..
         Informix::SecondLevelMonitor ( $sEnvFile,   # Running SLM..
                                        $sInformixDir,
                                        $iTillWhen ) : 1 ) ) { 
      #---------------------
      # Start-up confirmed..
      #---------------------
      &logInfo ( 20017,
      "<%s> Confirmed start-up of instance [%s] of Informix server",
      $sMyName, $sDBServerName );

      $iReturnCode = 1; # Success..
      last;
    }
    else {
      &logTrace ( 20018,
      "<%s> Could not confirm start-up. Sleeping for [%s] sec",
      $sMyName, $iSleepTime );

      $iReturnCode = 0; # Failure..
    }
  }

  #--------------------------------------------
  # Done for HP-UX and Solaris 10 local zones..
  #--------------------------------------------
  return $iReturnCode unless ( $bEnv );

  #---------------------------
  # Create IPC history files..
  #---------------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20019,
    "<%s> Insufficient time [%s] to create IPC history files",
    $sMyName, abs $iTimeout );
  }
  elsif ( Informix::CreateHistoryFiles ( O_IPCFILE,
                                         $sEnvFile,
                                         $sInformixDir,
                                         $iTimeout ) ) {
    &logTrace ( 20020,
    "<%s> Created IPC history files successfully",  $sMyName );
  }
  else {
    &logError ( 20021,
    "<%s> Could NOT create IPC history files successfully", $sMyName );
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool Stop ( IN String DBServerName,
                        IN String EnvFile,
                        IN String InformixDir );

 Description : Stop logic needed for offline entry pt

 Parameters :

 $sDBServerName         Unique name of this instance of the database server
 $sEnvFile              Absolute path to environment file to source
 $sInformixDir          Absolute path to installation directory
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Stop {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sDBServerName,
       $sEnvFile,
       $sInformixDir ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20022, "Subroutine <%s> called with:
  DBServerName             [%s]
  EnvFile                  [%s]
  InformixDir              [%s]",

  $sMyName, $sDBServerName, $sEnvFile, $sInformixDir );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + Informix::GetEngineTimeout ( 1 ); # With margin..

  #------------------------------------------
  # Declare stop command and args variables..
  #------------------------------------------
  my ( $sCommand, $sCommandArgs, $iTimeout, $iExitCode, $iSleepTime, $bFail,
       @lsCmdout, @liPids, @liShmIds ) = ();
  
  #------------------------------------
  # Define stop command and arguments..
  #------------------------------------
  $sCommand = "$sInformixDir/bin/onmode";
  $sCommandArgs = "-ky";

  &logTrace ( 20023,
  "<%s> Stopping Informix with Command [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #------------------------
  # Request shutdown..
  #------------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20024,
    "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = 0; # Failure..
  }
  elsif ( $iReturnCode =
          Informix::ExecLocalCmd ( "informix", 
                                   $sEnvFile, 
                                   $sCommand, 
                                   $sCommandArgs, 
                                   \$iExitCode, 
                                   $iTimeout, 
                                   \@lsCmdout ) ) {
    &logTrace ( 20025,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }
  else {
    ++ $bFail;
  }

  #----------------------------------------------
  # Remove persistent Informix system resources..
  #----------------------------------------------
  if ( $bFail ) { # Failure..
  
    &logInfo ( 20026,
    "<%s> Removing persistent system resources for Informix instance [%s]",
    $sMyName, $sDBServerName );

    #-----------------------------------------
    # Whack any persisting Informix processes..
    #-----------------------------------------

    &logTrace ( 20027,
    "<%s> Killing Informix processes for Informix instance [%s]",
    $sMyName, $sDBServerName );
  
    if ( Informix::KillInformixProcesses ( $sDBServerName,
                                           $sEnvFile,
                                           $sInformixDir ) ) {
      &logTrace ( 20028,
      "<%s> Processes for instance [%s] do not exist after shutdown",
      $sMyName, $sDBServerName );
    }
    else {
      &logError ( 20029,
      "<%s> Could not remove all processes for Informix instance [%s]. Exiting",
      $sMyName, $sDBServerName );
  
      $iReturnCode = 0; # Failure..
    }
    
    #----------------------------
    # Remove IPC resources here..
    #----------------------------
    if ( Informix::RemoveIPCs () ) {
      &logTrace ( 20030,
      "<%s> Confirmed IPC resources for Informix instance [%s] do not exist",
      $sMyName, $sDBServerName );
    }
    else {
      &logError ( 20031,
      "<%s> Could not confirm removal of IPC resources for Informix instance [%s]",
      $sMyName, $sDBServerName );
    }
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool RemoveIPCs ( void );

 Description : Requests removal of shared memory ids from history file.
               Deletes history file in any case.

 Parameters : None


 Returns : boolean value in the set of {0,1} where
           1 true     Could actually remove SHM ids.
           0 false    Anything else.
 -------------------------------------------------------------------------

=cut

sub RemoveIPCs {
  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20032, "Subroutine <%s> called to remove IPC resources",
  $sMyName );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iExitCode, $sResourceType, $sResourceName, $sIPCFile,
       @liShmIds ) = ();

  #----------------------------------------------
  # Derive our resource type and resource names..
  #----------------------------------------------
  $sResourceType = ( $0 =~ m@([^/]+)@g ) [-2];
  $sResourceName   = $ENV { VCS_LOG_RESOURCE_NAME };
  $sIPCFile = "/tmp/.VRTS$sResourceType/$sResourceName/IPC/onstat.out";

  #----------------------
  # Read in the SHM ids..
  #----------------------
  if ( open ( IPCFILE, $sIPCFile ) ) {

    if ( @liShmIds = map { m/(\d+)/ } <IPCFILE> ) {

      &logTrace ( 20033,
      "<%s> Got SHM Ids [%s] from IPC history file [%s]",
      $sMyName, join ( ' ', @liShmIds ), $sIPCFile );

      #-----------------------------
      # Request removal of SHM IDs..
      #-----------------------------
      if ( Proc::RemoveIPCResourcesForUser ( "root",
                                             O_IPCSHM,
                                             \@liShmIds ) ) {
        &logTrace ( 20034,
        "<%s> Confirmed SHM Ids [%s] do not exist on system",
        $sMyName, join ( ' ', @liShmIds ) );
      }
      else {
        &logError ( 20035,
        "<%s> Failed to remove all Shared Memories", $sMyName ); # Verify..
        $iReturnCode = 0; # Failure..
      }
    }
    else {
      &logWarn ( 20036,
      "<%s> Failed to get any SHM Id for removal from IPC history file [%s]",
      $sMyName, $sIPCFile );

      $iReturnCode = 0; # Failure..
    }

    #------------------------
    # Close IPC hitory file..
    #------------------------
    unless ( close ( IPCFILE ) ) {
      &logError ( 20037,
      "<%s> Could not close IPC history file [%s] after read: [%s]",
      $sMyName, $sIPCFile, $! );

      $iReturnCode = 0; # Failure..
    }

    #--------------------------
    # Remove IPC history file..
    #--------------------------
    if ( unlink ( $sIPCFile ) ) {
      &logTrace ( 20038,
      "<%s> Removed IPC history file [%s] successfully after read",
      $sMyName, $sIPCFile );
    }
    else {
      &logError ( 20039,
      "<%s> Could not remove IPC history file [%s] after read: [%s]",
      $sMyName, $sIPCFile, $! );

      $iReturnCode = 0; # Failure..
    }
  }
  else {
    &logTrace ( 20040,
    "<%s> Could not open IPC history file [%s] for read: [%s]",
    $sMyName, $sIPCFile, $! );

    $iReturnCode = 0; # Failure..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool Kill ( IN Integer CleanReason,
                        IN String DBServerName,
                        IN String EnvFile,
                        IN String InformixDir );

 Description : Forces shutdown of resource needed for clean entry point

 Parameters :

 $iCleanReason              Why clean was called
 $sDBServerName             Unique name of this instance of the database server
 $sEnvFile                  Absolute path to environment file to source
 $sInformixDir              Absolute path to installation directory
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Kill {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iCleanReason,
       $sDBServerName,
       $sEnvFile,
       $sInformixDir ) = @_;

  #-----------------------------
  # Figure out why were called..
  #-----------------------------
  my ( $sCleanReason, %hsCleanReasons ) = ();
  my @lsCleanStrings = qw ( Offline_timed_out    Offline_was_ineffective
                            Online_timed_out      Online_was_ineffective
                            Found_offline_unexpectedly Monitor_timed_out );

  @hsCleanReasons { 0 .. $#lsCleanStrings } = @lsCleanStrings; # Populate..
  ( $sCleanReason = $hsCleanReasons { $iCleanReason } ||       # Reason..
                    $iCleanReason ) =~ s/_/ /g;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20041, "Subroutine <%s> called with:
  CleanReason              [%s]
  DBServerName             [%s]
  EnvFile                  [%s]
  InformixDir              [%s]",

  $sMyName,     $sCleanReason,      $sDBServerName,      $sEnvFile,
  $sInformixDir );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + Informix::GetEngineTimeout ( 1 ); # With margin..

  #--------------------------
  # Declare local variables..
  #--------------------------
  my ( $sCommand, $sCommandArgs, $iTimeout, $bFail, $iExitCode,
       @lsFiles, @lsCmdout ) = ();
  
  #-------------------------------------
  # Define clean command and arguments..
  #-------------------------------------
  $sCommand = "$sInformixDir/bin/onmode";
  $sCommandArgs = "-ky";

  unless ( $hsCleanReasons { $iCleanReason } eq 'Offline_timed_out' ||
           $hsCleanReasons { $iCleanReason } eq 'Offline_was_ineffective' ) {
    #-------------------
    # Try to be gentle..
    #-------------------
    &logInfo ( 20042,
    "<%s> Attempting to stop Informix instance [%s] with Command [%s] Args [%s]",
    $sMyName, $sDBServerName, $sCommand, $sCommandArgs );
  
    #------------------------
    # Request shutdown..
    #------------------------
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
      &logError ( 20043,
      "<%s> Insufficient time to attempt normal shutdown. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );

      ++ $bFail; # Set fail flag..
    }
    elsif ( Informix::ExecLocalCmd ( "informix", 
                                     $sEnvFile, 
                                     $sCommand, 
                                     $sCommandArgs, 
                                     \$iExitCode, 
                                     $iTimeout, 
                                     \@lsCmdout ) ) {
      &logTrace ( 20044, 
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
    }
    else {
      ++ $bFail; # Set fail flag..
    }
  }
  else {
    ++ $bFail; # Set fail flag..
  }

  #-----------------------------------------
  # Whack any persisting Informix processes..
  #-----------------------------------------
  if ( $bFail ) {
    &logInfo ( 20045,
    "<%s> Killing Informix processes for instance [%s]", $sMyName, $sDBServerName );

    if ( Informix::KillInformixProcesses ( $sDBServerName,
                                           $sEnvFile,
                                           $sInformixDir ) ) {
      &logTrace ( 20046,
      "<%s> Processes for instance [%s] do not exist after shutdown",
      $sMyName, $sDBServerName );
    }
    else {
      &logError ( 20047,
      "<%s> Could not remove all processes for instance [%s]. Exiting",
      $sMyName, $sDBServerName );

      $iReturnCode = 0; # Failure..
    }

    #----------------------------
    # Remove IPC resources here..
    #----------------------------
    if ( Informix::RemoveIPCs () ) {
      &logTrace ( 20048,
      "<%s> Confirmed IPC resources for Informix instance [%s] do not exist", 
      $sMyName, $sDBServerName );
    }
    else {
      &logError ( 20049,
      "<%s> Could not confirm removal of IPC resources for Informix instance [%s]",
      $sMyName, $sDBServerName );
    }
  }
  
  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : int GetResourceState ( IN String DBServerName,
                                   IN String EnvFile,
                                   IN String InformixDir );

 Description : First level monitor to determine state of resource.

 Parameters :

 $sDBServerName             Unique name of this instance of the database server
 $sEnvFile                  Absolute path to environment file to source
 $sInformixDir              Absolute path to installation directory
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetResourceState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sDBServerName,
       $sEnvFile,
       $sInformixDir ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20050, "Subroutine <%s> called:
  DBServerName             [%s]
  EnvFile                  [%s]
  InformixDir              [%s]",

  $sMyName, $sDBServerName, $sEnvFile, $sInformixDir );

  my @liDummyPids = ();
  my $iReturnCode = Informix::GetProcState ( \@liDummyPids,
                                             $sDBServerName,
                                             $sEnvFile,
                                             $sInformixDir,
                                             1 ); # Use FFLM..

  if ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    &logTrace ( 20051,
    "<%s> Found all required processes. Return code set to [%s]",
    $sMyName, $iReturnCode );
  }
  else {
    &logTrace ( 20052,
    "<%s> Did not find all required processes. Returning [%s]",
    $sMyName, $iReturnCode );
  }

  #--------------------------------
  # Populate the global PID array..
  #--------------------------------
  @liRunningProcs = @liDummyPids;
                                                                                
  &logTrace ( 20053,
  "<%s> Populating global running proclist with [%s] PIDs [%s]. Returning [%s]",
  $sMyName, scalar @liRunningProcs, join ( ' ', @liRunningProcs ),
  $iReturnCode );

  return ( $iReturnCode );
}


=pod

 -------------------------------------------------------------------------
 Function : int GetProcState ( OUT ListReference OnlinePids,
                               IN String DBServerName,
                               IN String EnvFile,
                               IN String InformixDir,
                               IN Integer UseFFLM );

 Description : First level monitor to determine state of resource.

 Parameters :

 $rliOnlinePids             PIDs of relevant running processes
 $sDBServerName             Unique name of this instance of the database server
 $sEnvFile                  Absolute path to environment file to source
 $sInformixDir              Absolute path to installation directory
 $iUseFFLM                  Filtering procedure
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetProcState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $rliOnlinePids,
       $sDBServerName,
       $sEnvFile,
       $sInformixDir,
       $iUseFFLM ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20054, "Subroutine <%s> called with:
  OnlinePids               [%s]
  DBServerName             [%s]
  EnvFile                  [%s]
  InformixDir              [%s]
  UseFFLM                  [%s]",

  $sMyName,     $rliOnlinePids,     $sDBServerName,      $sEnvFile,
  $sInformixDir, $iUseFFLM );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #----------------------
  # Initialise Variables..
  #----------------------
  my ( $iProcessCount, $iUseRFLM, $iFound, $iUid, $iPid,
       $iParent, $iFChild, $iOninitPID, $sLang, $sKey,
       $rhlProc, $sPidFile, $sProc, $sTemp, $sOSVer,
       @hProcListHash, @lsUpdates, @lsValues,
       %hiOninitPids, %hiProcPids, %hsSrvrDetails, %hsCopy ) = ();

  @$rliOnlinePids = (); # Truncate..

  #--------------------------------------------------
  # Attempt to get environment of running processes..
  #--------------------------------------------------
  $bEnv = 0 if ( $^O eq 'solaris' || $^O eq 'hpux' );
  
  &logTrace ( 20057,
  "<%s> Environment flag set to [%s] on OS [%s]", 
  $sMyName, $bEnv, $^O );
  
  #------------------------------------
  # Attempt to get monitoring details..
  #------------------------------------
  if ( ( 0 == ( $iUid = getpwnam "informix" ) ) && # User is the super-user..
       ( 2 == $iUseFFLM ) ) {                      # And called to kill-by-UID..
    &logTrace ( 20058, "<%s> Unsafe to kill processes by User [%s] UID [%s].
    Using filter-based approach.", $sMyName, "informix", $iUid ),
    $iUseFFLM = 1;
  }

  if ( $iUseFFLM != 2 ) { # Called for first level check..
    if ( Informix::GetSrvrDetails ( $sDBServerName, 
                                    $sEnvFile,
                                    $sInformixDir,
                                    \%hsSrvrDetails ) ) {
      &logTrace ( 20059,
      "<%s> Got details for monitoring Informix instance [%s]",
      $sMyName, $sDBServerName );
  
      %hsCopy = %hsSrvrDetails;        # Backup..
      $iUseRFLM = scalar keys %hsCopy; # Set to entire count..
    }
    else {
      &logError ( 20060,
      "<%s> Failed to get details for monitoring Informix instance [%s]",
      $sMyName, $sDBServerName );
      return VCSDef_MonitorExit::UNKNOWN;
    }
                                  
    #---------------------------------
    # Check if we attempt to be fast.. 
    #---------------------------------
    if ( $iUseFFLM ) { # FFLM..
      $iUseRFLM = 0;   # Hopefully skip..

      foreach $sKey ( keys %hsCopy ) {          # Use copy..
        @lsUpdates = ();                        # Reset..
        @lsValues = @{ $hsCopy { $sKey } };     # Get values..
        $sTemp .= "\nInformix process [$sKey]\n"; # Display string..

        #----------------
        # Read in pairs..
        #----------------
        while ( ( $sPidFile, $sProc ) = splice ( @lsValues, 0, 2 ) ) {

          #-----------
          # Run FFLM..
          #-----------
          $iFound = Informix::FFLMStatus ( $sPidFile,
                                           $sProc,
                                           $iUid,
                                           \$iPid );

          #----------------
          # Update status..
          #----------------
          push ( @lsUpdates, $sPidFile, $sProc, $iFound );

          $sTemp .= "PID File     [$sPidFile]\n";
          $sTemp .= "Srch Process [$sProc]\n";
          $sTemp .= "FFLM Status  [$iFound]\n\n";
        }

        @{ $hsCopy { $sKey } } = @lsUpdates;

        #------------------------
        # Handle monitor status..
        #------------------------
        if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
          push ( @$rliOnlinePids, $iPid ); # Push..
        }
        else {
          ++ $iUseRFLM; # Mark it for RFLM..
        }
      }

      #----------------
      # Print summary..
      #----------------
      &logTrace ( 20061,
      "<%s> Informix instance [%s] process details ---%s\n\n",
      $sMyName, $sDBServerName, $sTemp );

      ( $sTemp, $sPidFile, $sProc ) = (); # Release memory..
    }
    else { # Use RLFM..
      #-----------------------------------------
      # Mark status of each service as offline..
      #-----------------------------------------
      $iFound = VCSDef_MonitorExit::OFFLINE; # Set to offline..
      foreach $sKey ( keys %hsCopy ) {
        @lsUpdates = (); # Reset..

        @lsValues = @{ $hsCopy { $sKey } }; # Get values..
        while ( scalar @lsValues ) {
          push ( @lsUpdates, splice ( @lsValues, 0, 2 ), $iFound );
        }

        @{ $hsCopy { $sKey } } = @lsUpdates; # Update..
      }
    }
  }
  else { # Called for Kill..
    &logTrace ( 20062,
    "<%s> Called to get processes for User [%s] UID [%s]",
    $sMyName, "informix", $iUid );
    ++ $iUseRFLM;
  }

  #-----------------------
  # Regular slow :-( FLM..
  #-----------------------
  if ( ! $bEnv && $iUseRFLM ) { # HP-UX..

    #------------------------------------------------------------------
    # For HP-UX/Solaris local zone, check if onstat utility was found..
    #------------------------------------------------------------------
    if ( ! $bEnv        &&      # On HP-UX/Solaris local zone..
         ! $bOnlineNode &&      # Steady state offline..
         ! -d $sInformixDir ) { # And InformixDir absent..
      &logTrace ( 20063,
      "<%s> Skipping onstat check on OFFLINE node. Returning OFFLINE.",
      $sMyName );
      return VCSDef_MonitorExit::OFFLINE;
    }
    else {
      &logTrace ( 20064,
      "<%s> Running onstat based first level monitor", $sMyName );
    }

    #-------------------------
    # Set LANG=C for parsing..
    #-------------------------
    $sLang = $ENV { LANG }; # Backup..
    $ENV { LANG } = 'C';    # Now set..

    if ( $iPid = Informix::CreateHistoryFiles ( O_HISTORY,
                                                $sEnvFile,
                                                $sInformixDir,
                                                5 ) ) { # Hard code timeout to 5 sec..
      &logTrace ( 20065,
      "<%s> Found Informix instance [%s] online with PID [%s].
       Created IPC history files successfully",
       $sMyName, $sDBServerName, $iPid );

      ${ $hsCopy { oninit } }[-1] = VCSDef_MonitorExit::ONLINE;
      @$rliOnlinePids = $iPid; # Store only the parent's PID..
    }
    else {
     &logTrace ( 20066,
     "<%s> Could NOT create IPC history files successfully", $sMyName );
    }

    #----------
    # Restore..
    #----------
    defined $sLang ? $ENV { LANG } = $sLang:delete $ENV { LANG }; # Restore..

  }
  elsif ( $bEnv && $iUseRFLM ) { # Not on HP-UX and Solaris 10 local zone..
    &logTrace ( 20067,
    "<%s> Running regular first level monitor", $sMyName );

    #----------------------------------
    # Set LANG for better performance..
    #----------------------------------
    $sLang = $ENV { LANG }; # Backup..
    $ENV { LANG } = 'C';    # Now set..

    #-----------------------------------
    # Call processes and store in hash..
    #-----------------------------------
    $iProcessCount = Proc::GetLongProcessListHash ( \@hProcListHash,
                                                    $bEnv );

    #----------
    # Restore..
    #----------
    defined $sLang ? $ENV { LANG } = $sLang:delete $ENV { LANG }; # Restore..

    #----------------------
    # Check if we got any..
    #----------------------
    if ( $iProcessCount ) {
      &logTrace ( 20068,
      "<%s> Got [%s] processes from proc table for filtering",
      $sMyName, $iProcessCount );
    }
    else {
      &logError ( 20069,
      "<%s> Call to get process list did not return any processes",
      $sMyName );
      return VCSDef_MonitorExit::OFFLINE; # Offline..
    }

    $iParent = 1; # Set to init..

    #-----------------------
    # Search for processes..
    #-----------------------
    foreach $rhlProc ( @hProcListHash ) {
      #--------------------------------------------------------------
      # Each $rhlProc is a ref. to a hash with the following keys..
      # "UID"     =>  <user id>
      # "PID"     =>  <process id>
      # "PPID"    =>  <parent process id>
      # "COMMAND" =>  <command string>
      #--------------------------------------------------------------

      next unless defined $$rhlProc { UID }; # Bug in Proc::GetLongProcessListHash..
      next if $$rhlProc { PID } == $$;       # Skip self..
      $iPid = $$rhlProc { PID };             # Assign..

      &logTrace ( 20070,
      "<%s> Searching PID [%s] PPID [%s] UID [%s] CMD [%s]",
      $sMyName, $$rhlProc { PID }, $$rhlProc { PPID }, $$rhlProc { UID },
      $$rhlProc { COMMAND } );

      if ( $$rhlProc { UID } == $iUid || # Process owner is "informix"..
           0 == $$rhlProc { UID } ) {    # Or process owner is "root"..
        #-----------------------------------------------
        # Look for processes that were offline in FFLM..
        #-----------------------------------------------
        if ( $iUseFFLM != 2 ) { # Request to filter..
          foreach $sKey ( keys %hsCopy ) {
            @lsUpdates = (); # Reset..
            @lsValues = @{ $hsCopy { $sKey } }; # Get values..
    
            #--------------------------------------------------
            # Should have details per three elements of array..
            # <PIDFILE> <PROCESS> <STATE>
            #--------------------------------------------------
            while ( ( $sPidFile, $sProc, $iFound ) =
                      splice ( @lsValues, 0, 3 ) ) {
              if ( $$rhlProc { COMMAND } =~ /$sProc/ ) {      # Now found..
                $hiOninitPids { $iPid } = $$rhlProc { PPID }; # Mark..
  
                &logTrace ( 20071,
                "<%s> Found PID [%s] PPID [%s] CMD [%s] with filter [%s]",
                $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND },
                $sProc );

                #--------------------------------
                # Mark the parent oninit process..
                #--------------------------------
                if ( $$rhlProc { PPID } == $iParent ) {        # Parent is init..
                  $hiProcPids { $sKey } = $iOninitPID = $iPid; # Store parent oninit..
                  $iFound = VCSDef_MonitorExit::ONLINE;        # Mark online..

                  &logTrace ( 20072,
                  "<%s> +--- Found parent oninit process with PID [%s] PPID [%s] ---+",
                  $sMyName, $iOninitPID, $$rhlProc { PPID } );
                }
              }

              #----------------
              # Update status..
              #----------------
              push ( @lsUpdates, $sPidFile, $sProc, $iFound );
            }
    
            @{ $hsCopy { $sKey } } = @lsUpdates; # Update..
          }
        }
      }
    }

    #----------------------------------
    # Mark all oninit in current zone..
    #----------------------------------
    if ( $iOninitPID ) { # Parent oninit found..
      foreach $iPid ( keys %hiOninitPids ) {
        if ( $hiOninitPids { $iPid } == $iOninitPID ) {
          $iFChild = $iPid; # "Agraj", the first born..
          last;
        }
      }

      &logTrace ( 20073,
      "<%s> Got Parent oninit PID [%s] with First Child [%s]",
      $sMyName, $iOninitPID, $iFChild );

      @$rliOnlinePids =                   # Store PIDs..
      grep {                              # After filtering only those..
        $hiOninitPids { $_ } == $iFChild; # Whose PPID is "Agraj"..
      } keys %hiOninitPids;               # From all found..

      #--------------------------------
      # Insert the parent and "Agraj"..
      #--------------------------------
      unshift ( @$rliOnlinePids, $iOninitPID, $iFChild ); 

      &logTrace ( 20074,
      "<%s> Located PIDs [%s]. Total [%s] PIDs found for Informix instance [%s]",
      $sMyName, join ( ' ', @$rliOnlinePids ), scalar @$rliOnlinePids,
      $sDBServerName );
    }
    else {
      &logTrace ( 20075,
      "<%s> Failed to find parent oninit process for instance [%s]",
      $sMyName, $sDBServerName ); 
    }
  }

  #------------------------------------
  # Set final status for each process..
  #------------------------------------
  my ( $iOnline, $iOffline, $sEntryPointName ) = ();

  ( $sEntryPointName = $0 ) =~ s#.*/##; # Basename..
  $sTemp = ''; # Ensure..

  foreach $sKey ( keys %hsCopy ) {
    @lsValues = @{ $hsCopy { $sKey } }; # Get values..
    $sTemp .= "\nProcess      [$sKey]\n";

    while ( ( $sPidFile, $sProc, $iFound ) =
              splice ( @lsValues, 0, 3 ) ) {
      #-----------------------------------
      # Create PID files for use in FFLM..
      #-----------------------------------
      if ( $iUseRFLM && $bEnv                        &&      # In RFLM and not HP-UX..
           $iUseFFLM != 2                            &&      # Asked to filter..
           $hiProcPids { $sKey }                     &&      # PID populated..
           ( $sEntryPointName =~ /^(monitor|online)$/o ) ) { # In monitor or online..

        &logTrace ( 20076,
        "<%s> Creating PID file [%s] with PID [%s]",
        $sMyName, $sPidFile, $hiProcPids { $sKey } );

        #---------
        # Create..
        #---------
        if ( Informix::CreatePidFile ( $sPidFile,
                                       $hiProcPids { $sKey } ) ) {
          &logTrace ( 20077,
          "<%s> Successfully created PID file [%s]", $sMyName, $sPidFile );
        }
        else {
          &logError ( 20078,
          "<%s> Could not create PID file [%s] successfully",
          $sMyName, $sPidFile );
        }
      }

      #-------------
      # Set status..
      #-------------
      if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
        $iOnline = $iFound;
      }
      elsif ( VCSDef_MonitorExit::OFFLINE == $iFound ) {
        $iOffline = $iFound;
      }

      $sTemp .= "Srch Process [$sProc]\n";
      $sTemp .= "FLM Status   [$iFound]\n\n";
    }

    $sTemp .= "--------------------";
  }

  #-------------------
  # Set final status..
  #-------------------
  $iReturnCode = $iOffline || $iOnline || VCSDef_MonitorExit::UNKNOWN;

  #------------------------------------------------
  # Reset status based on PIDs, if not in monitor..
  #------------------------------------------------
  if ( $sEntryPointName ne 'monitor'                   &&
       ( VCSDef_MonitorExit::OFFLINE == $iReturnCode ) &&
       scalar @$rliOnlinePids ) {
    $iReturnCode = VCSDef_MonitorExit::PARTIAL_ONLINE;

    &logTrace ( 20079,
    "<%s> Found [%s] PIDs [%s] for instance [%s] of Informix.
    Resetting return code to [%s]",
    $sMyName, scalar @$rliOnlinePids, join ( ' ', @$rliOnlinePids ),
    $sDBServerName, $iReturnCode );
  }

  #-----------------------
  # Report on FLM status..
  #-----------------------
  if ( $iUseFFLM != 2 ) { # Request to filter..
    &logTrace ( 20080,
    "<%s> Informix process details ---\n%s\nReturn code is [%s]",
    $sMyName, $sTemp, $iReturnCode );
    $sTemp = undef; # Release..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int Monitor ( IN String EnvFile,
                          IN String InformixDir,
                          IN String ServerType,
                          IN String MonitorProgram,
                          IN Integer SecondLevelMonitor );

 Description: Performs the in-depth monitoring

 Parameters :

 $sEnvFile                  Absolute path to environment file to source
 $sInformixDir              Absolute path to installation directory
 $sServerType               Standard/HDR/SharedDisk
 $sMonitorProgram           Full path to external monitor program
 $iSecondLevelMonitor       Cycle count to run in-depth monitoring
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if in-depth monitoring cannot be completed.
 -------------------------------------------------------------------------

=cut

sub Monitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sInformixDir,
       $sDBServerName, 
       $sServerType,
       $sMonitorProgram,
       $iSecondLevelMonitor ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20081, "Subroutine <%s> called with:
  EnvFile                  [%s]
  InformixDir              [%s]
  DBServerName             [%s]
  ServerType               [%s]
  MonitorProgram           [%s]
  SecondLevelMonitor       [%s]",

  $sMyName,  $sEnvFile,  $sInformixDir, $sDBServerName, 
  $sServerType, $sMonitorProgram, $iSecondLevelMonitor );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  my $iTillWhen = undef;

  # If Online Check if SDS Primary
  if ( $sServerType =~ /SharedDisk/ ){
     my $bIsPrimary = IsPrimary ( $sEnvFile, $sInformixDir, $sDBServerName );
     CreateSDSPrimFile ( $sInformixDir, $sDBServerName, $bIsPrimary );
  }

  #----------------------------------------------
  # In-depth monitoring, if correct time to run..
  #----------------------------------------------
  if ( VCSagentFW::RunSecondLevelMonitor ( $iSecondLevelMonitor ) ) {
    #--------------------------
    # Just In Time Validation..
    #--------------------------
    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         Informix::DelayedAttributeCheck ( $sEnvFile,
                                           $sInformixDir ) ) ) {
      &logTrace ( 20082,
      "<%s> Delayed attribute check completed successfully",
      $sMyName );
    }
    else {
      &logError ( 20083,
      "<%s> Attribute validation failed. Returning [%s]",
      $sMyName, $iReturnCode );
      return $iReturnCode;
    }

    &logTrace ( 20084,
    "<%s> Proceeding with second level monitoring", $sMyName );

    #-------------------------------------------------------
    # Upper bound..
    # Time in EPOCH when entry pt started + VCS time limit..
    #-------------------------------------------------------
    $iTillWhen = $^T + Informix::GetEngineTimeout ( 1 );

    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         Informix::SecondLevelMonitor ( $sEnvFile,
                                        $sInformixDir,
                                        $iTillWhen ) ) ) {

      &logTrace ( 20085,
      "<%s> Informix instance online with status [%s] after SLM",
      $sMyName, $iReturnCode );
    }
    else {
      &logError ( 20086,
      "<%s> Informix instance NOT online after SLM. Status [%s]",
      $sMyName, $iReturnCode );
      return $iReturnCode;
    }
  }

  #------------------------------------------------
  # Exit if external monitor has not been defined..
  #------------------------------------------------
  return $iReturnCode unless defined $sMonitorProgram;

  #---------------------------------------
  # MonitorProgram defined..
  # Separate the executable and the args..
  #---------------------------------------
  my ( $sExtCmd, $sExtCmdArgs ) = $sMonitorProgram # Separate
                                =~ /(\S+)\s*(.*)/; # Cmd and args..

  $sExtCmdArgs = '' unless defined $sExtCmdArgs;   # Prevent undefs..

  if ( $sExtCmd ) {                                # Defined..
    #----------------------------------------
    # Get time bearings, if not got earlier..
    #----------------------------------------
    $iTillWhen ||= $^T + Informix::GetEngineTimeout ( 1 );

    if ( VCSDef_MonitorExit::ONLINE ==             # and found online..
         ( $iReturnCode =                          # by the status..
         Informix::ExternalMonitor ( $sEnvFile,
                                     $sExtCmd,
                                     $sExtCmdArgs,
                                     $iTillWhen ) ) ) {
      &logTrace ( 20087,
      "<%s> Informix instance found online with status [%s]
      after executing external monitor [%s] with args [%s]",
      $sMyName, $iReturnCode, $sExtCmd, $sExtCmdArgs );

    }
    else {
      &logError ( 20088,
      "<%s> Informix instance NOT online using external monitor
      [%s] with args [%s]. Status [%s]",
      $sMyName, $sExtCmd, $sExtCmdArgs, $iReturnCode );
    }
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int DelayedAttributeCheck ( IN String EnvFile,
                                        IN String InformixDir );

 Description : Just In Time Validation

 Parameters :

 $sEnvFile                  Absolute path to environment file to source
 $sInformixDir              Absolute path to installation directory
 
 Returns  : Integer value as defined below:

 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub DelayedAttributeCheck {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sInformixDir ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20089, "Subroutine <%s> called with:
  EnvFile                  [%s]
  InformixDir              [%s]",

  $sMyName, $sEnvFile, $sInformixDir );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  my @lsArgs = qw ( EnvFile InformixDir );
  my @lsReqdArgs = qw ( EnvFile InformixDir );
  my ( @lsMissing, %hsArgList ) = ();

  @hsArgList { @lsArgs } = @_; # Assign..

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20090,
    "<%s> The following mandatory fields were not set [%s], Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #--------------------
  # Return, if failed..
  #--------------------
  if ( VCSDef_MonitorExit::ONLINE != $iReturnCode ) {
    return $iReturnCode;
  }

  #------------------------------
  # Validate parameter EnvFile..
  #------------------------------
  if ( ! -e $sEnvFile ) {
    &logError ( 20091, "<%s> EnvFile [%s] file does not exist",
    $sMyName, $sEnvFile );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #----------------------------------
  # Validate parameter InformixDir..
  #----------------------------------
  if ( ! -d $sInformixDir ) {
    &logError ( 20092, "<%s> InformixDir [%s] directory does not exist",
    $sMyName, $sInformixDir );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int SecondLevelMonitor ( IN String EnvFile,
                                     IN String InformixDir,
                                     IN Integer TillWhen );

 Description : Does in-depth monitoring, if enabled.

 Parameters :

 $sEnvFile                  Absolute path to environment file to source
 $sInformixDir              Absolute path to installation directory
 $iTillWhen                 Time in EPOCH left for IInd level monitoring
 
 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if the command cannot be completed.
 -------------------------------------------------------------------------

=cut

sub SecondLevelMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sInformixDir,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20093, "Subroutine <%s> called with:
  EnvFile                  [%s]
  InformixDir              [%s]
  TillWhen                 [%s]",

  $sMyName, $sEnvFile, $sInformixDir, $iTillWhen );  

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iTimeout, $iExitCode, $sCommand, $sCommandArgs,
       $sOutput, $sState, $sMode, @lsCmdout ) = ();
  
  my %hiCodes = ( 2 => "Sec", 5 => "Prim", 2 => "SDS" );

  #---------------------------------------------
  # Declare monitor command and args variables..
  #---------------------------------------------
  $sCommand = "$sInformixDir/bin/onstat";
  $sCommandArgs = "-";

  #-------------------------------
  # Validate and execute command..
  #-------------------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20094,
    "<%s> Insufficient time to continue in-depth monitoring.
    Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    return VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  &logTrace ( 20095,
  "<%s> Executing Cmd [%s] with args [%s] within [%s] sec",
  $sMyName, $sCommand, $sCommandArgs, $iTimeout );

  #---------------------
  # Ignore return code..
  #---------------------
  Informix::ExecLocalCmd ( "informix", 
                            $sEnvFile, 
                            $sCommand, 
                            $sCommandArgs, 
                            \$iExitCode, 
                            $iTimeout, 
                            \@lsCmdout );

  #--------------------------------------
  # Check exit code of command directly..
  #--------------------------------------
  if ( ! defined $iExitCode ) {
    $lsCmdout [0] ||= 'undef'; # Set if not defined..
    $iReturnCode = $lsCmdout [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
                                               VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  elsif ( 5 == $iExitCode || 2 == $iExitCode ) {
    #--------------------
    # Get output string..
    #--------------------
    ( $sOutput ) = grep { /\S/ } @lsCmdout;

    #-----------------------------------
    # Check if HDR has been configured..
    #-----------------------------------
    if ( ( $sState, $sMode ) = $sOutput
      =~ m/--\s+(.*)\s+\((.*?)\)/ ) { # Configured..
      &logTrace ( 20096,
      "<%s> Found database in State [%s] acting as [%s]",
      $sMyName, $sState, $sMode );

      #--------------------------
      # Set state of database..
      #--------------------------
      if ( defined $hiCodes { $iExitCode } &&
           $hiCodes { $iExitCode } eq $sMode ) {
        $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..
      }
      else {
        $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Unknown..
      }
    }
    elsif ( 5 == $iExitCode ) { # Regular..
      $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..
    }
  }
  elsif ( 255 == $iExitCode ) {
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  else {
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Unknown state..
  }

  &logTrace ( 20097,
  "<%s> Returning status [%s]", $sMyName, $iReturnCode );

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int ExternalMonitor ( IN String EnvFile,
                                  IN String Cmd,
                                  IN String CmdArgs,
                                  IN Integer TillWhen );

 Description : Handles user defined external monitor program

 Parameters :

 $sEnvFile                  Absolute path to environment file to source
 $sCmd                      User defined utilty to execute
 $sCmdArgs                  Command line arguments
 $iTillWhen                 Time in EPOCH left for IInd level monitoring
 
 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if the command cannot be completed.
 -------------------------------------------------------------------------

=cut

sub ExternalMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sCmd,
       $sCmdArgs,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20098, "Subroutine <%s> called with:
  EnvFile                  [%s]
  Cmd                      [%s]
  CmdArgs                  [%s]
  TillWhen                 [%s]",

  $sMyName, $sEnvFile, $sCmd, $sCmdArgs, $iTillWhen );

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #------------------------
  # Setup local variables..
  #------------------------
  my ( $iExitCode, $iTimeout, @lsCmdout ) = ();
  my $iBuf = 5; # Secs..

  #-------------------------
  # Handled exit code hash..
  #-------------------------
  my %hiExitCodes = (  0, VCSDef_MonitorExit::ONLINE,
                     110, VCSDef_MonitorExit::ONLINE,
                       1, VCSDef_MonitorExit::OFFLINE,
                     100, VCSDef_MonitorExit::OFFLINE,
                          VCSDef_MonitorExit::TIMEOUT,
                          VCSDef_MonitorExit::TIMEOUT );

  #---------------------------------------------
  # Check if command can be found and executed..
  #---------------------------------------------
  if ( ! -e $sCmd ) {
   &logError ( 20099,
    "<%s> External monitor [%s] does not exist",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ! -x $sCmd ) {
   &logError ( 20100,
    "<%s> External monitor [%s] not executable.",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( $sEnvFile =~ /\S/ && ! -e $sEnvFile ) {
    &logError ( 20101,
    "<%s> EnvFile [%s] does not exist", $sMyName, $sEnvFile );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ( $iTimeout = $iTillWhen - ( time + $iBuf ) ) <= 5 ) { # Lack of time..
   &logError ( 20102,
    "<%s> Insufficient time to proceed with external monitor [%s].
    Short by [%s] sec", $sMyName, $sCmd, abs ( $iTimeout ) );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }

  #--------------------------------------------
  # Return if validation failed at this point..
  #--------------------------------------------
  unless ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    return $iReturnCode;
  }

  &logTrace ( 20103,
  "<%s> Executing external monitor with Command [%s] Args [%s] in [%s] secs",
  $sMyName, $sCmd, $sCmdArgs, $iTimeout );

  #----------------------
  # Execute the command..
  #----------------------
  Informix::ExecLocalCmd ( "informix", 
                           $sEnvFile, 
                           $sCmd, 
                           $sCmdArgs, 
                           \$iExitCode, 
                           $iTimeout, 
                           \@lsCmdout );
                               
  #------------------------
  # Report on exit status..
  #------------------------
  if ( ! defined $iExitCode ) {
    $lsCmdout [0] ||= 'undef'; # Set if not defined..
    $iExitCode = $lsCmdout [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
                                             VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  elsif ( $hiExitCodes { $iExitCode } &&
          VCSDef_MonitorExit::ONLINE ==
          $hiExitCodes { $iExitCode } ) {
    &logTrace ( 20104,
    "<%s> Command [%s] with args [%s] executed successfully with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
  }
  else {
    &logWarn ( 20105,
    "<%s> Command [%s] with args [%s] execution failed with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
  }

  #--------------------------------
  # Setup appropriate return code..
  #--------------------------------
  $iReturnCode = $hiExitCodes { $iExitCode } ||
                 VCSDef_MonitorExit::UNKNOWN;

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool CreatePidFile ( IN String PidFile,
                                 In Integer Pid );


 Description : Creates PID file with given PID

 Parameters :

 $sPidFile     File to create
 $iPid         PID to store

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub CreatePidFile {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $iPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20106, "Subroutine <%s> called with:
  PidFile         [%s]
  Pid             [%s]",

  $sMyName, $sPidFile, $iPid );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-----------------------
  # Check against undefs..
  #-----------------------
  unless ( defined $iPid && $iPid =~ /^\d+$/o) {
    $iPid ||= 'undef'; # Prevent undefs.
    &logError ( 20107, "<%s> PID [%s] is not a valid positive number",
    $sMyName, $iPid );
    return 0; # Proceed no further..
  }

  #------------------
  # Create PID file..
  #------------------
  &logTrace ( 20108,
  "<%s> Creating PID file [%s] with PID [%s]",
  $sMyName, $sPidFile, $iPid );

  if ( sysopen ( PIDFILE, $sPidFile, O_WRONLY|O_SYNC|O_CREAT|O_TRUNC,
                 0600 ) ) {
    print PIDFILE $iPid, "\n"; # Update..

    unless ( close ( PIDFILE ) ) {
      &logWarn ( 20109,
      "<%s> Failed to close PidFile [%s] after update: [%s]",
      $sMyName, $sPidFile, $! );
      $iReturnCode = 0;
    }
  }
  else {
    &logError ( 20110,
    "<%s> Failed to create PID file [%s] with PID [%s]: [%s]",
    $sMyName, $sPidFile, $iPid, $! );
    $iReturnCode = 0;
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int FFLMStatus ( IN String PidFile,
                             IN String Proc,
                             IN Integer Uid,
                             OUT RefInt Pid );

 Description: Checks status of Process listed in PidFile, if running

 Parameters :

 $sPidFile       Absolute path to PID file of service
 $sProc          Process string to search for in proc table
 $iUid           UNIX UID of owner
 $riPid          Reference to existing PID

 Returns : Integer value as defined below:
 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 -------------------------------------------------------------------------

=cut

sub FFLMStatus {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $sProc,
       $iUid,
       $riPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20111, "Subroutine <%s> called with:
  PidFile                [%s]
  Proc                   [%s]
  UID                    [%s]
  riPid                  [%s]",

  $sMyName, $sPidFile, $sProc, $iUid, $riPid );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..
  $$riPid = 0; # Reset..

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iPid, $rhlProc, @lProcessHash ) = ();

  #----------------------------------
  # Start by testing the attributes..
  #----------------------------------
  if ( $sProc !~ /\S/ ) {
    &logError ( 20112, "<%s> Called with empty [%s] process string",
    $sMyName, $sProc );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -e $sPidFile ) {
    &logTrace ( 20113, "<%s> PID file [%s] does not exist",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -s $sPidFile ) {
    &logWarn ( 20114, "<%s> PID file [%s] is of zero bytes",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #---------------------
  # Return, if offline..
  #---------------------
  return $iReturnCode unless VCSDef_MonitorExit::ONLINE == $iReturnCode;
 
  if ( open ( PIDFILE, $sPidFile ) ) {
    ( $iPid ) = grep { ( s/\n// || 1 ) && # Chomp..
                         s/.*?(\d+)$/$1/  # Get PID..
                } <PIDFILE>;

    $iPid = '' unless defined $iPid;      # Avoid undefs..

    close ( PIDFILE ) ||                  # Close file..
    &logWarn ( 20115,
    "<%s> Could not close PID file [%s] after read [%s]",
    $sMyName, $sPidFile, $! );
  }
  else {
    &logError ( 20116,
    "<%s> Failed to open PID file [%s] for read: [%s]",
    $sMyName, $sPidFile, $! );
    return VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #------------------------
  # Check if PID is valid..
  #------------------------
  if ( $iPid =~ /\d+/ &&     # If the valid number..
       kill ( 0, $iPid ) ) { # is a PID that's alive..

    &logTrace ( 20117,
    "<%s> PID [%s] from PID file [%s] exists",
    $sMyName, $iPid, $sPidFile );

    #-----------------------
    # Search in proc table..
    #-----------------------
    if ( Informix::GetPidDetails ( $iPid,
                                   \@lProcessHash ) ) {
      #--------------
      # Got details..
      #--------------
      foreach $rhlProc ( @lProcessHash ) {
        &logTrace ( 20118,
        "<%s> Process details: UID [%s] PID [%s] PPID [%s] ARGS [%s]",
        $sMyName, $$rhlProc { UID }, $$rhlProc { PID }, $$rhlProc { PPID },
        $$rhlProc { COMMAND } );

        #--------------
        # Now compare..
        #--------------
        if ( ( $$rhlProc { UID } == $iUid  ||        # Owner match..
               0 == $$rhlProc { UID } ) &&           # Owner match..
               $$rhlProc { COMMAND } =~ /$sProc/ ) { # Filter match..
          &logTrace ( 20119,
          "<%s> +--- PID [%s] PPID [%s] owned by UID [%s] with ARGS [%s] contains [%s] ---+",
          $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { UID },
          $$rhlProc { COMMAND }, $sProc );
          $$riPid = $iPid; # Store..
        }
        else {
          &logTrace ( 20120,
          "<%s> Ignoring PID [%s] PPID [%s] owned by UID [%s] with ARGS [%s] using regex [%s]",
          $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { UID },
          $$rhlProc { COMMAND }, $sProc );
          $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
        }
      }
    }
    else {
      &logTrace ( 20121,
      "<%s> Could not get details for PID [%s]",
      $sMyName, $iPid );
      $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
    }
  }
  else {
    &logTrace ( 20122, "<%s> PID [$iPid] does not exist",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  @lProcessHash = (); # Truncate..
  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool KillInformixProcesses ( IN String DBServerName,
                                         IN String EnvFile,
                                         IN String InformixDir );

 Description  : Kills processes for this instance of Informix
 
 Parameters :

 $sDBServerName           Unique name of this instance of the database server
 $sEnvFile                Absolute path to environment file to source
 $sInformixDir            Absolute path to installation directory
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------
=cut

sub KillInformixProcesses {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sDBServerName,
       $sEnvFile,
       $sInformixDir ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20123, "Subroutine <%s> called with:
  DBServerName             [%s]
  EnvFile                  [%s]
  InformixDir              [%s]",

  $sMyName, $sDBServerName, $sEnvFile, $sInformixDir );
 
  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #-------------------------
  # Set up local variables..
  #-------------------------
  my ( @liGoners, @liUnDead ) = ();

  #---------------------------------------------
  # Get list of processes running for this Uid..
  #---------------------------------------------
  if ( ( caller 1 )[3] eq 'Informix::Start' ) { # During online..
    @liGoners = @liRunningProcs;                # Get PIDs to kill..
  }
  else {
    Informix::GetProcState ( \@liGoners,
                             $sDBServerName,
                             $sEnvFile,
                             $sInformixDir,
                             ! $bEnv ); # Set filtering choice..
  }

  #------------------------------------------------------------------
  # We do a two level kill. The SIGINQ ensures  that  the  PID  is  a
  # valid process before we actually issue the SIGKILL. We don't want
  # to report a failure for SIGKILL in case  of  defunct  or  missing
  # processes.
  #------------------------------------------------------------------
  if ( scalar @liGoners ) {

    unless ( $bEnv ) { # We get only one PID..
      &logTrace ( 20124,
      "<%s> Got PID [%s] to kill along with its process tree",
      $sMyName, $liGoners [ 0 ] );

      Proc::KillPIDandAllChildren ( $liGoners [ 0 ] ); # No return code..
    }
    else {
      #----------------------
      # Failed to kill some..
      #----------------------
      if ( @liUnDead = grep {
                           kill ( 0, $_ ) && # Valid..
                         ! kill ( 9, $_ )    # and could NOT be killed..
                       } @liGoners ) {       # From what's found..
        &logError ( 20125,
        "<%s> PIDs not killed [%s] Total not killed [%s]",
        $sMyName, join ( ' ', @liUnDead ), scalar @liUnDead );
        $iReturnCode = 0; # Failure..
      }
      else { # All killed..
        &logInfo ( 20126,
        "<%s> Killed PIDs [%s] Total killed: [%s]",
        $sMyName, join ( ' ', @liGoners ), scalar @liGoners );
      }
    }
  }
  else {
    &logTrace ( 20127,
    "<%s> Could not locate any process for instance [%s] of Informix",
    $sMyName, $sDBServerName );
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool ExecLocalCmd ( IN String User,
                                IN String EnvFile,
                                IN String Command,
                                IN String CommandArgs,
                                OUT RefInt ExitCode,
                                IN Integer Timeout,
                                OUT ListReference Cmdout );

 Description : Wrapper function for
               Sys::RunWithEnvCmdWithOutputWithTimeOut to collect output
               of the command.
 
 Parameters :

 $sUser            UNIX user to run the command as
 $sEnvFile       Absolute path to location of EnvFile to source
 $sCommand         Absolute path to location of command
 $sCommandArgs     Arguments to command
 $riExitCode       Reference to collect exit code of command
 $iTimeout         Time limit to execute the command
 $rlsCmdout        Stores o/p from command if successful
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ExecLocalCmd {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sUser,
       $sEnvFile,
       $sCommand,
       $sCommandArgs,
       $riExitCode,
       $iTimeout,
       $rlsCmdout ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20128, "Subroutine <%s> called with:
  User        [%s]
  EnvFile     [%s]
  Command     [%s]
  CommandArgs [%s]
  riExitCode  [%s]
  Timeout     [%s]
  rlsCmdout   [%s]",

  $sMyName,
  $sUser, $sEnvFile, $sCommand, $sCommandArgs, $riExitCode,
  $iTimeout, $rlsCmdout );

  #---------------------------------------
  # Setup return code to hopeful success..
  #---------------------------------------
  my $iReturnCode = 1;

  #-------------------
  # Set up arguments..
  #-------------------
  my $bSourceUserEnv = 0; # Do not source..
  my $bTimedOut = 0;      # Timeout variable..

  @$rlsCmdout = (); # Reset..

  #-----------
  # Validate..
  #-----------
  if ( ! -e $sCommand ) {
    &logError ( 20129,
    "<%s> Cmd [%s] does not exist", $sMyName, $sCommand );
    $iReturnCode = 0; # Failure..
  }
  elsif ( ! -x $sCommand ) {
    &logError ( 20130,
    "<%s> Cmd [%s] is not executable", $sMyName, $sCommand );
    $iReturnCode = 0; # Failure..
  }
  elsif ( $iTimeout <= 1 ) {
    &logError ( 20131,
    "<%s> Insufficient time [%s] to execute Cmd [%s] with Args [%s].",
    $sMyName, $iTimeout, $sCommand, $sCommandArgs );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return if failed.. 

  #----------------------
  # Set LOGNAME to user..
  #----------------------
  my $sOldLogName = $ENV { LOGNAME }; # Backup..
  $ENV { LOGNAME } = $sUser;          # Now set..

  &logTrace ( 20132,
  "<%s> Executing Command [%s] with Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  $iReturnCode = Sys::RunWithEnvCmdWithOutputWithTimeOut ( $sEnvFile,
                                                           $bSourceUserEnv,
                                                           $sCommand,
                                                           $sCommandArgs,
                                                           $riExitCode,
                                                           $sUser,
                                                           $rlsCmdout,
                                                           \$bTimedOut,
                                                           $iTimeout );

  #---------------------------
  # Remove trailing newlines..
  #---------------------------
  chomp @$rlsCmdout;

  #--------------------
  # Check return code..
  #--------------------
  if ( $iReturnCode ) {   # No error with command execution..
    if ( $$riExitCode ) { # Failure..
      &logTrace ( 20133,
      "<%s> Command [%s] with args [%s] returned exit code [%s]",
      $sMyName, $sCommand, $sCommandArgs, $$riExitCode );
      $iReturnCode = 0; # Failure..
    }
    else {
      &logTrace ( 20134,
      "<%s> Command [%s] with args [%s] executed successfully with exit code [%s]",
      $sMyName, $sCommand, $sCommandArgs, $$riExitCode );
    }
  }
  else { # Execution error..
    @$rlsCmdout = (); # Reset..

    if ( $bTimedOut ) {  # We timed out..
      &logError ( 20135,
      "<%s> Command [%s] with args [%s] did not complete within [%s] seconds",
      $sMyName, $sCommand, $sCommandArgs, $iTimeout );
      $$rlsCmdout [0] = 'timeout'; # Timeout..
    }
    else { # Failure due to something else..
      &logError ( 20136,
      "<%s> Failed in call to RunWithEnvCmdWithOutputWithTimeOut",
      $sMyName );
    }
  }

  #----------
  # Restore..
  #----------
  defined $sOldLogName ? $ENV { LOGNAME } = $sOldLogName:delete $ENV { LOGNAME };

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool GetPidDetails ( IN Integer Pid,
                                 OUT RefList ProcList );


 Description : Gets long process details for a known Pid if running

 Parameters :

 $iPid         PID whose details are requested
 $rlProcList   Ref to List to store details
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetPidDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iPid,
       $rlProcList ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20137, "Subroutine <%s> called with:
  Pid             [%s]
  rlProcList      [%s]",

  $sMyName, $iPid, $rlProcList );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #---------------
  # Validate PID..
  #---------------
  unless ( defined $iPid    &&   # Catch undefs..
           $iPid =~ /^\d+$/ &&   # And invalid PIDs..
           kill ( 0, $iPid ) ) { # And defunct ones..
    &logError ( 20138,
    "<%s> PID [%s] is invalid", $sMyName, $iPid );
    return 0;
  }

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sEnv, $sPsCmd, $sPsCmdArgs, $sLang, $sFmt, $sUnix95, $sSwitch, $iExitCode,
       %hsProc ) = ();

  $sSwitch = $^O;    # Get OS..
  @$rlProcList = (); # Truncate..

  #-------------
  # Store LANG..
  #-------------
  $sLang = $ENV { LANG };
  $ENV { LANG } = 'C';

  #---------------------------
  # Switch to the correct OS..
  #---------------------------
  goto $sSwitch; {
    solaris:

      $sEnv = $bEnv ? 'e' : ''; # Env on Global zone only..
      $sPsCmd = "/usr/ucb/ps";
      $sPsCmdArgs = "axww${sEnv}l $iPid";
      $sFmt = '[\da-z]+\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';

    last;

    hpux:

      #-----------------------
      # Force XPG4 behaviour..
      #-----------------------
      $sUnix95 = $ENV { UNIX95 };
      $ENV { UNIX95 } = 'XPG4';

      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "-p $iPid -xxlo uid,pid,ppid,args";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';

    last;

    aix:

      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "axwwel $iPid";
      $sFmt =  '[\da-e]+\s+[A-Z]\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';

    last;


    linux:

      $sPsCmd = "/bin/ps";
      $sPsCmdArgs = "wweo uid,pid,ppid,args $iPid";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';

    last;
  }

  &logTrace ( 20139, "<%s> OS [%s] PsCmd [%s] PsCmdArgs [%s]",
  $sMyName, $sSwitch, $sPsCmd, $sPsCmdArgs );

  #-------------------
  # Validate command..
  #-------------------
  if ( ! -x $sPsCmd ) {
    &logError ( 20140,
    "<%s> ps(1b) command [%s] is not executable", $sMyName, $sPsCmd );
    return 0; # Failure..
  }

  #-------------------------------
  # Fire command and get details..
  #-------------------------------
  @hsProc { qw ( UID PID PPID COMMAND ) } =
  map {
    s/\n//;                             # Chomp..
    s/^\s+//;                           # Strip leading spaces..
    /^$sFmt$/o;                         # Extract fields..
  } readpipe ( "$sPsCmd $sPsCmdArgs" ); # From the command..

  $iExitCode = $? >> 8; # Get exit code..

  #------------------------
  # Restore env variables..
  #------------------------
  defined $sLang ? $ENV { LANG } = $sLang:delete $ENV { LANG };
  defined $sUnix95 ? $ENV { UNIX95 } = $sUnix95:delete $ENV { UNIX95 };

  #------------------
  # Check exit code..
  #------------------
  if ( $iExitCode ) {
    &logError ( 20141,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
    $sMyName, $sPsCmd, $sPsCmdArgs, $iExitCode );

    %hsProc = (); # Truncate..
    return 0;
  }

  #---------------------
  # Fill up the bucket..
  #---------------------
  push ( @$rlProcList, \%hsProc ); # Update..

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : bool GetSrvrDetails ( IN String DBServerName,
                                  IN String EnvFile,
                                  IN String InformixDir,
                                  OUT RefHashList SrvrDetails );

 Description : Gives PID file and process string for given server type

 Parameters  :

 $sDBServerName             Unique name of this instance of the database server
 $sEnvFile                  Absolute path to environment file to source
 $sInformixDir              Absolute path to installation directory
 $rhlSrvrDetails            Stores details of PID file and process string
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetSrvrDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sDBServerName,
       $sEnvFile,
       $sInformixDir,
       $rhlSrvrDetails ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20142, "Subroutine <%s> called with:
  DBServerName             [%s]
  EnvFile                  [%s]
  InformixDir              [%s]
  SrvrDetails              [%s]",

  $sMyName, $sDBServerName, $sEnvFile, $sInformixDir, $rhlSrvrDetails );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = 1; # Set to true..

  #-------------------
  # Define variables..
  #-------------------
  my ( $sKey, $sProc, $sPidFile, $sPIDDir, $sResourceType,
       $sResourceName, $iPat ) = ();

  #----------------------------------------------
  # Derive our resource type and resource names..
  #----------------------------------------------
  $sResourceType = ( $0 =~ m@([^/]+)@g ) [-2];
  $sResourceName   = $ENV { VCS_LOG_RESOURCE_NAME };
  $sPIDDir = "/tmp/.VRTS$sResourceType/$sResourceName/PID";

  #------------------------------
  # Create PID folder if needed..
  #------------------------------
  if ( -d $sPIDDir ) {
    &logTrace ( 20143, "<%s> PID folder [%s] exists",
    $sMyName, $sPIDDir );
  }
  elsif ( ! mkpath ( $sPIDDir, 0, 0700 ) ) {
    &logWarn ( 20144, "<%s> Failed to create PID folder [%s]",
    $sMyName, $sPIDDir );
    return 0; # Cannot proceed..
  }

  %$rhlSrvrDetails = (); # Truncate..

  #---------------------------------------------
  # Set appropriate filter string and PID file..
  #---------------------------------------------
  $sProc = $bEnv ? qr!\boninit\b.*\bINFORMIXSERVER=$sDBServerName\b! :
                   qr!\boninit\b!;

  $sKey = 'oninit';
  $sPidFile = "$sPIDDir/pidfile.$sKey";
  @{ $$rhlSrvrDetails { $sKey } } = ( $sPidFile, $sProc ); # Store..
 
  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int GetEngineTimeout ( IN Bool UseMargin );

 Description : Calculates engine timeout for entry point
 
 Parameters  :

 $bUseMargin      Flag to indicate if safety margin is requested

                  boolean value in the set of {0,1} where
                  0 No safety margin needed
                  1 Safety margin to be used

 Returns:  Time in secs.
 -------------------------------------------------------------------------

=cut

sub GetEngineTimeout {
  #----------------
  # Get Arguments..
  #----------------
  my ( $bUseMargin ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20145, "Subroutine <%s> called with:
  UseMargin  [%s]",
  $sMyName, $bUseMargin );

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sAttribute, $sCommand, $sCommandArgs,
       $iEngineTimeout, $iExitCode ) = ();

  #----------------------------------
  # Grace margins from VCS timeouts..
  #----------------------------------
  my %hiGraceMargin = ( 'MonitorTimeout', 10,   # Seconds..
                        'OnlineTimeout',  10,
                        'OfflineTimeout', 10,
                        'CleanTimeout',   20 );

  #--------------------------------------------------
  # Attempt to get resource level overridden values..
  #--------------------------------------------------
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  my ( $sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;
  ( $sAttribute = $sEntryPointName ) =~ s/(.*)/\u$1Timeout/;

  #-----------------------------------
  # Query if it has already been set..
  #-----------------------------------
  if ( defined $::giEPTimeout &&    # Defined..
       $::giEPTimeout =~ /^\d+/ ) { # to an integer..
    $iEngineTimeout = $::giEPTimeout;

    &logTrace ( 20146,
    "<%s> Using AEPTimeout [%s] as EP timeout value",
    $sMyName, $iEngineTimeout );
  }
  else {
    #-----------------------------------------
    # Form query string based on entry point..
    #-----------------------------------------
    $sCommand = $sCmdHares; # Get global..
    $sCommandArgs = "-value $sResourceName $sAttribute";
  
    #----------------------------------
    # Fire command and collect output..
    #----------------------------------
    ( $iEngineTimeout = readpipe ( "$sCommand $sCommandArgs 2>/dev/null" ) )
                      =~ s/\n//;
  
    #--------------------
    # Check exit status..
    #--------------------
    if ( $iExitCode = $? >> 8 ) {
      &logTrace ( 20147,
      "<%s> Command [%s] with Args [%s] failed with exitcode [%s].",
      $sMyName, $sCommand, $sCommandArgs, $iExitCode );
    }
    elsif ( ! $iEngineTimeout ) { # Undefined value..
      &logTrace ( 20148, "<%s> Could not get [%s] from engine",
      $sMyName, $sAttribute );
    }
    else { # Success..
      &logTrace ( 20149, "<%s> Got [%s] as [%s] sec from engine",
      $sMyName, $sAttribute, $iEngineTimeout );
    }
  
    #-----------------------------------------
    # Form query string based on entry point..
    #-----------------------------------------
    unless ( $iEngineTimeout ) {
      $sCommand = $sCmdHatype; # Get global..
      ( $sAttribute = $sEntryPointName ) =~ s/(.*)/\u$1Timeout/;
      $sResourceType = "Informix" if $sResourceType !~ /\w/; # Ensure..
      $sCommandArgs = "-value $sResourceType $sAttribute";
    
      #----------------------------------
      # Fire command and collect output..
      #----------------------------------
      ( $iEngineTimeout = readpipe ( "$sCommand $sCommandArgs" ) ) =~ s/\n//;
    
      #--------------------
      # Check exit status..
      #--------------------
      if ( $iExitCode = $? >> 8 ) {
        &logTrace ( 20150,
        "<%s> Command [%s] with Args [%s] failed with exitcode [%s]",
        $sMyName, $sCommand, $sCommandArgs, $iExitCode );
        $iEngineTimeout = $bUseMargin ? $hiGraceMargin { $sAttribute } : 0;
      }
      elsif ( ! $iEngineTimeout ) { # Empty value..
        &logTrace ( 20151, "<%s> Could not get [%s] from engine",
        $sMyName, $sAttribute );
        $iEngineTimeout = $bUseMargin ? $hiGraceMargin { $sAttribute } : 0;
      }
      else { # Success..
        &logTrace ( 20152, "<%s> Got [%s] as [%s] sec from engine",
        $sMyName, $sAttribute, $iEngineTimeout );
      }
    }
    else {
        &logTrace ( 20153, "<%s> Got [%s] as [%s] sec from engine",
        $sMyName, $sAttribute, $iEngineTimeout );
    }
  }

  #----------------------------
  # Apply margin if requested.. 
  #----------------------------
  if ( $bUseMargin     &&                                    # Margin requested..
       $iEngineTimeout >= $hiGraceMargin { $sAttribute } ) { # And valid..
    $iEngineTimeout -= $hiGraceMargin { $sAttribute };       # Apply..

    &logTrace ( 20154,
    "<%s> Setting up [%s] sec margin. Returning [%s]",
    $sMyName, $hiGraceMargin { $sAttribute }, $iEngineTimeout );
  }

  return ( $iEngineTimeout ); # Secs..
}

=pod

 --------------------------------------------------------------------------
 Function : Integer CreateHistoryFiles ( IN Integer FileFlag,
                                         IN String EnvFile,
                                         IN String InformixDir,
                                         IN Integer Timeout );

 Description : Creates history files based on request.

 Parameters :

 $iFileFlag        Behaviour flag
 $sEnvFile         Absolute path to environment file to source
 $sInformixDir     Absolute path to installation directory
 $iTimeout         Time limit to execute the command

 Values for Behaviour flag:

 O_PIDFILE          : Create PID file only
 O_IPCFILE          : Create SHM file only
 O_HISTORY          : Create both PID and SHM files

 Return Value: integer value in the set of {0,n} where
               0 false
               n PID of the parent oninit process
 --------------------------------------------------------------------------

=cut

sub CreateHistoryFiles {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iFileFlag,
       $sEnvFile,
       $sInformixDir,
       $iTimeout ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20155, "Subroutine <%s> called with:
  FileFlag           [%s]
  EnvFile            [%s]
  InformixDir        [%s]
  Timeout            [%s]",
  $sMyName, $iFileFlag, $sEnvFile, $sInformixDir, $iTimeout );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sResourceType, $sResourceName, $sPIDFile, $sIPCFile,
       $sIPCDir, $sCmd, $sCmdArgs, $i, $iPid, $iExitCode,
       @liShmIds, @lsCmdout ) = ();

  #----------------------------------------------
  # Derive our resource type and resource names..
  #----------------------------------------------
  $sResourceType = ( $0 =~ m@([^/]+)@g ) [-2];
  $sResourceName   = $ENV { VCS_LOG_RESOURCE_NAME };
  
  $sCmd = "$sInformixDir/bin/onstat";
  $iPid = 0; # Default to 0..

  #------------ Do not indent ---------
  $sCmdArgs = "-i << EOF
glo
seg
q
EOF";
  #------------ Do not indent ---------

  if ( $iReturnCode = Informix::ExecLocalCmd ( "informix",
                                               $sEnvFile,
                                               $sCmd,
                                               $sCmdArgs,
                                               \$iExitCode,
                                               $iTimeout,
                                               \@lsCmdout ) ) {
    if ( $lsCmdout [0] =~
         /WARNING: IBM Informix Dynamic Server is no longer running/ ) {

      $iReturnCode = 0; # Failure..

      &logTrace ( 20156,
      "<%s> Cmd [%s] with args [%s] reported the following output [%s].
      Returning [%s]\n",
      $sMyName, $sCmd, $sCmdArgs, $lsCmdout [0], $iReturnCode );

      return $iReturnCode; # Offline..
    }
    else {
      #---------------
      # Parse output..
      #---------------
      ( $iPid, @liShmIds ) =
      grep {
        ( s/\n// || 1 )                  && # Chomp and make true..
        ( s/^\s//g || 1 )                && # Strip leading whites..
        ( s/.*\s(\d+)\s+cpu.*/$1/ )      || # Get the PID..
        ( s/^id\b.*// && ++ $i && /\S/ ) || # Mark the SHM output, set flag..
        ( $i && s/^(\d+).*/$1/ )            # If flag, Get IDs..
      } @lsCmdout;                          # From output..
  
      #--------------------------------------------
      # Double check if reported PID is really up..
      #--------------------------------------------
      if ( kill ( 0, $iPid ) ) {
        &logTrace ( 20157,
        "<%s> PID of parent oninit process [%s]
        SHM IDs [%s]", $sMyName, $iPid, join ( ' ', @liShmIds ) );
      }
      else {
        $iReturnCode = 0; # Failure..

        &logWarn ( 20158,
        "<%s> PID of parent oninit process [%s] does not exist.
        Returning [%s]", $sMyName, $iPid, $iReturnCode );
      }
    }
  }
  else { # Failed in execution..
    &logTrace ( 20159,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode );

    return $iReturnCode;
  }

  #-----------------------------
  # Request to create PID file..
  #-----------------------------
  if ( $iFileFlag & O_PIDFILE ) {

    $sPIDFile = "/tmp/.VRTS$sResourceType/$sResourceName/PID/pidfile.oninit";

    #---------
    # Create..
    #---------
    if ( $iReturnCode = Informix::CreatePidFile ( $sPIDFile,
                                                  $iPid ) ) {
      &logTrace ( 20160,
      "<%s> Successfully created PID file [%s]", $sMyName, $sPIDFile );
    }
    else {
      &logError ( 20161,
      "<%s> Could not create PID file [%s] successfully",
      $sMyName, $sPIDFile );
    }
  }

  #-----------------------------
  # Request to create IPC file..
  #-----------------------------
  if ( $iFileFlag & O_IPCFILE ) {
    $sIPCDir = "/tmp/.VRTS$sResourceType/$sResourceName/IPC";
    $sIPCFile = "$sIPCDir/onstat.out";

    #------------------------------
    # Create IPC folder if needed..
    #------------------------------
    if ( -d $sIPCDir ) {
      &logTrace ( 20162, "<%s> IPC folder [%s] exists",
      $sMyName, $sIPCDir );
    }
    elsif ( ! mkpath ( $sIPCDir, 0, 0700 ) ) {
      &logWarn ( 20163, "<%s> Failed to create IPC folder [%s]",
      $sMyName, $sIPCDir );
      return 0; # Cannot proceed..
    }

    #------------------
    # Create IPC file..
    #------------------
    &logTrace ( 20164,
    "<%s> Creating IPC history file [%s] with SHM IDs [%s]",
    $sMyName, $sIPCFile, join ( ' ', @liShmIds ) );

    if ( sysopen ( IPCFILE, $sIPCFile, O_WRONLY|O_SYNC|O_CREAT|O_TRUNC,
                   0600 ) ) {
      print IPCFILE join ( chr 012, @liShmIds ), "\n"; # Update..
  
      unless ( close ( IPCFILE ) ) {
        &logError ( 20165,
        "<%s> Failed to close IPC file [%s] after update: [%s]",
        $sMyName, $sIPCFile, $! );
        return 0;
      }
    }
    else {
      &logError ( 20166,
      "<%s> Failed to create IPC file [%s]: [%s]",
      $sMyName, $sIPCFile, $! );
      return 0;
    }
  }

  return $iPid;
}

=pod
#---------------------------------------------------------------
 Function : bool SwitchPrim ( IN String OldPrimDB,
                              IN String DBServerName,
                              IN String EnvFile,
                              IN String InformixDir );

 Description : Promoting SDS Secondary to Primary

 Parameters :

 $sOldPrimDB            Faulted DBServerName
 $sDBServerName		Database Server Name
 $sEnvFile              Environment File
 $sInformixDir          Informix Home dir


 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
#---------------------------------------------------------------
=cut

sub SwitchPrim {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sOldPrimDB,
       $sDBServerName,
       $sEnvFile,
       $sInformixDir ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20184, "Subroutine <%s> called with:
  OldPrimaryDB		   [%s]
  DBServerName             [%s]
  EnvFile                  [%s]
  InformixDir              [%s]",

  $sMyName, $sOldPrimDB, $sDBServerName, $sEnvFile, $sInformixDir);

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sAttribute, $sCommand, $sCommandArgs,
       @lsCmdout, $iTimeout, $iExitCode ) = ();

  $iTimeout = 100;
  #------------------------------------------
  # Define switch command and args variables..
  #------------------------------------------
  $sCommand = "$sInformixDir/bin/onmode";
  $sCommandArgs = "-d set SDS primary $sDBServerName force";

  #-------------------
  # Request switching
  #-------------------
  if ( $iReturnCode =
    Informix::ExecLocalCmd ( "informix",
                             $sEnvFile,
                             $sCommand,
                             $sCommandArgs,
                             \$iExitCode,
                             $iTimeout,
                             \@lsCmdout ) ) {
    &logTrace ( 20185,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }
  else {
    &logTrace ( 20186,
    "<%s> Unable to Promote SD Secondary [%s] to SD Primary.
          Cmd [%s] with args [%s] failed with output[%s]",
    $sMyName, $sDBServerName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );

    $iReturnCode = 0; # Failure..
  }
  
  my $sSDSFile = "$sInformixDir/.VRTSInfxSDPrimDB";
  if ( $iReturnCode ) {  # If action is successful 
      #---------------------------------
      # Check if it is really primary.
      # If Yes then update the SDS File
      #---------------------------------
      if ( IsPrimary ( $sEnvFile, $sInformixDir, $sDBServerName ) ){

         my @lsData = ();
         #----------------------
         # Read in the SDS file..
         #----------------------
         if ( sysopen ( SDSFILE, $sSDSFile, O_RDONLY ) ) {
           @lsData = <SDSFILE>;
         
           #------------------------
           # Close file..
           #------------------------
           unless ( close ( SDSFILE ) ) {
             &logTrace ( 20187,
             "<%s> Could not close SDS file [%s] after read: [%s]",
             $sMyName, $sSDSFile, $! );
           }
         }
         else {
           &logError ( 20188,
           "<%s> Failed to open SDSfile [%s] for read : [%s]",
           $sMyName, $sSDSFile, $! );
           $iReturnCode = 0;
         }
         
         #-------------------------------------------------
         # Replace the old Primary DB Server Name with New
         #-------------------------------------------------
         grep { s/$sOldPrimDB/$sDBServerName/ } @lsData;
         
         #----------------------
         # Modify the SDS file..
         #----------------------
         if ( sysopen ( SDSFILE, $sSDSFile, O_WRONLY|O_TRUNC ) ) {
           print SDSFILE @lsData;

           #------------------------
           # Close file..
           #------------------------
           if ( close ( SDSFILE ) ) {
             &logTrace ( 20249,
             "<%s> Successfully updated the SDS file [%s]",
             $sMyName, $sSDSFile );
           }
           else {
             &logError ( 20189,
             "<%s> Could not close SDS file [%s] after write: [%s]",
             $sMyName, $sSDSFile, $! );
           }
         }
         else {
           &logError ( 20190,
           "<%s> Failed to open SDSfile [%s] for write: [%s]",
           $sMyName, $sSDSFile, $! );
           $iReturnCode = 0;
         }
      }
  }

  if ( $iReturnCode ) {
     &logInfo ( 20242,
     "<%s> Successfully Promoted the SD Secondary [%s] to SD Primary.",
     $sMyName, $sDBServerName );
  }
  return $iReturnCode;
}

=pod

#---------------------------------------------------------------
 Function : bool PreOnlineTrigger( IN String Sys,
                                   IN String Grp,
                                   IN String Reason,
                                   IN String FaultedNode );

 Description : Pre-online trigger for Informix

 Parameters :

 $sSys                System where the group is asked to be made online
 $sGrp                Group for which preonline has been invoked
 $sReason             Contains the string MANUAL or FAULT
 $sFaultedNode        Cluster node where the fault occurred due to FAULT

 Returns : boolean value in the set of {0,1} where

           0 false
           1 true
#---------------------------------------------------------------

=cut

sub PreOnlineTrigger{
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSys,
       $sGrp,
       $sReason,
       $sFaultedNode ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20191, "Subroutine <%s> called with:
  Sys             [%s]
  Grp             [%s]
  Reason          [%s]
  FaultedNode     [%s]",
  $sMyName, $sSys, $sGrp, $sReason, $sFaultedNode );

  #-------------------------------------------
  # Set return code optimistically to online..
  #-------------------------------------------
  my $iReturnCode = 1; # Success..

  #-------------------------
  # Define local variables..
  #-------------------------
  my $sCmd = $sCmdHares; # Get global..
  my %hsZones = map { $_ => 1 } qw ( 5.10 5.11 ); # For Solaris zones..
  my ( $sOSVer, $sCmdArgs, @lsHaresOut, $iExitCode ) = ();

  #---------------------
  # Validate arguments..
  #---------------------
  if ( $sReason eq 'MANUAL' ) {
    &logTrace ( 20192,
    "<%s> Invoked during a normal switch-over or online.", 
    $sMyName );
  }
  elsif ( $sSys =~ /\S/ && ! inet_aton ( $sSys ) ) {
    &logError ( 20193,
    "<%s> Destination cluster node [%s] cannot be resolved",
    $sMyName, $sSys );

    $iReturnCode = 0; # Error..
  }
  elsif ( $sFaultedNode ne 'UNDEF' && $sFaultedNode =~ /\S/ && ! inet_aton ( $sFaultedNode ) ) {
    &logError ( 20194,
    "<%s> Faulted cluster node [%s] cannot be resolved",
    $sMyName, $sFaultedNode );

    $iReturnCode = 0; # Error..
  }
  elsif ( ! -e $sCmd ) {
    &logError ( 20195,
    "<%s> Command [%s] does not exist", 
    $sMyName, $sCmd );

    $iReturnCode = 0; # Error..
  }
  elsif ( ! -x $sCmd ) {
    &logError ( 20196,
    "<%s> Command [%s] is not executable", 
    $sMyName, $sCmd );

    $iReturnCode = 0; # Error..
  }

  $iReturnCode || return $iReturnCode; # Return, if failed..

  #---------------------------------------------------
  # Verify group is really faulted if Reason is FAULT.
  #---------------------------------------------------
  if ( $sReason ne "MANUAL" ) {
    $sCmd = $sCmdHares;
    $sCmdArgs = "-list Group=$sGrp State=FAULTED";

    @lsHaresOut = grep {
                    ( s/\n// || 1 ) &&
                    s/\s+.*//
                  } readpipe ( "$sCmd $sCmdArgs" );

    #-------------------
    # Verify exit code..
    #-------------------
    if ( $iExitCode = $? >> 8 ) {
      &logTrace ( 20199,
      "<%s> Cmd [%s] with args [%s] failed with exit code [%s]. Group [%s] does not have any faulted resources.",
      $sMyName, $sCmd, $sCmdArgs, $iExitCode, $sGrp );

      #-----------------------------
      # Check for node FAULT event..
      #-----------------------------
      if ( $sFaultedNode eq 'UNDEF' && $sReason eq 'FAULT' ) { 
        &logWarn ( 20200,
        "<%s> Missing faulted node name. Possible node FAULT event detected", 
        $sMyName );
      }
      elsif ( $hsZones { $^O eq 'solaris' ? ( uname ) [2] : '' } ) { # Handle Solaris zones..
        #---------------------------------------------------------------
        # Add zone specific support for Enqueue when zone itself faults.
        #---------------------------------------------------------------
        &logInfo ( 20202,
        "<%s> Checking if the Group [%s] is configured for Solaris non-global zones",
        $sMyName, $sGrp );
  
        my $sOut = ();
        my @lsOut = ();
        my $iCount = ();
  
        $sCmdArgs = "-display -attribute ContainerName -group $sGrp -type Informix";
        @lsHaresOut = grep { (s/\n// || 1) } readpipe ( "$sCmd $sCmdArgs" );
  
        shift @lsHaresOut;
  
        foreach $sOut (@lsHaresOut) {
          @lsOut = split( '\s+', $sOut);
          $iCount = scalar @lsOut;
          if( (4 == $iCount) && ($lsOut[2] =~ /$sSys/ || $lsOut[2] =~ /global/) ) {
            &logWarn ( 20203,
            "<%s> ContainerName attribute is configured. Possible zone [%s] FAULT event detected", 
            $sMyName, $lsOut[3] );
  
            last;
          }
        }
      }
    }
    else {
      &logTrace ( 20204,
      "<%s> Group [%s] has [%s] resource(s) in FAULTED state",
      $sMyName, $sGrp, join ( ' ', @lsHaresOut ) );
    }
  }
  elsif ( $sFaultedNode eq 'UNDEF' && $sReason eq 'MANUAL' ) {
    &logTrace ( 20250,
    "<%s> Manual Switch is performed for the Group [%s] to System [%s].",
    $sMyName, $sGrp, $sSys );
  }

  #----------------
  # Re-initialise..
  #----------------
  $sCmdArgs = "-display -type Informix -attribute " .
              "State Group InformixDir ServerType DBServerName SDSInstances";

  #---------------------------------------------
  # Assume the group name is valid and proceed..
  #---------------------------------------------
  @lsHaresOut = readpipe ( "$sCmd $sCmdArgs" );

  #-----------------------------------------
  # Check if command executed successfully..
  #-----------------------------------------
  if ( $iExitCode = $? >> 8 ) {
    &logError ( 20205,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode );

    $iReturnCode = 0;
  }
  elsif ( $lsHaresOut[0] && $lsHaresOut[0] =~ /10554.*No resource exists with/o ) {
    &logError ( 20206,
    "<%s> Cmd [%s] with args [%s] failed with reason [%s]",
    $sMyName, $sCmd, $sCmdArgs,  $lsHaresOut[0] );

    $iReturnCode = 0;
  }
  $iReturnCode || return $iReturnCode; # Retrun if Error..

  push @lsHaresOut, "#"; # Insert trailing de-marker..
 
  my ( @lsCpy, @lsEnq, @lsNext ,$i, $sRes, $sGroup,
       $sState, @lsFaultedSys, $sSystem ) = ();

  @lsCpy = @lsHaresOut;
  #--------------------------------
  # Check if FAULT or SWITCH Event 
  #--------------------------------
  if ( $sReason eq 'FAULT') {
    if ( $sFaultedNode ){
      @lsFaultedSys = ( $sFaultedNode );
    }
    else {
      @lsFaultedSys =
      grep {
        ( s/\n// || 1 )                 && # Chomp, and always make true..
        ( s/.*\s(\S+)\s+FAULTED$/$1/  )    # If FAULTED, get node name, set flag..
      }@lsCpy;
    }
    $sState = "FAULTED";

    &logTrace ( 20251,
    "<%s> Fault event - Faulted Systems are [%s].",
    $sMyName, join (',', @lsFaultedSys) );
  }
  elsif ( $sReason eq 'MANUAL'){
    @lsFaultedSys =
    grep {
      ( s/\n// || 1 )                   && # Chomp, and always make true..
      ( s/.*\s(\S+)\s+OFFLINE$/$1/  )      # If OFFLINE, get node name, set flag..
    }@lsCpy;
    $sState = "OFFLINE";

    &logTrace ( 20252,
    "<%s> Manual Switch over - Offline Systems are [%s].",
    $sMyName, join (',', @lsFaultedSys) );
  }

  my %hUniq = ();
  @lsFaultedSys = grep { !$hUniq{$_}++ && $_ !~ $sSys } @lsFaultedSys; # Remove Duplicate and target node

# -- Foreach FAULTED/OFFLINE node --
  foreach $sSystem (@lsFaultedSys){
    @lsCpy = @lsHaresOut;  # Backup..
    @lsEnq =
    grep {
      ( s/\n// || 1 )                                    && # Chomp, and always make true..
      ( /^(\S+).*Group.*\s+(\S+)/                        && # Mark resource, and group name..
      ( $sGroup = "$1:$2" ) && 0 )                       || # Store, and always make false..
      ( s/.*\s$sSystem\s+$sState$/$sSystem/ && ++ $i )   || # If faulted, get node name, set flag..
      ( $i                                               && # If flag..
      /ServerType/                                       && # Store ServerType
      s/.*\s+// )                                        || # 
      ( $i                                               && # If flag..
      /DBServerName/                                     && # Store DBServerName
      s/.*\s+// )                                        || # 
      ( $i                                               && # If flag..
      /InformixDir/                                      && # Store InformixDir
      s/.*\s+// )                                        || # 
      ( $i                                               && # If flag..
      /SDSInstances/                                     && # Store SDSInstances
      s/.*\s+global\s+(.*)/$1/ )                         || # 
      ( $i                                               && # If flag, use demarker..
      s/#/sprintf ( $sGroup, -- $i )/e )                    # to print the group name.. reset flag..
    }@lsCpy;                                                # From the copy..

  #-------------------
  # Start processing..
  #-------------------

  my ($sPrimaryDB, $sInformixDir, $sNodeName, $sSDSFile,
      $sDBServerName, $sServerType, $sSDSInstances, 
      $sResGroup, $bPrimFaulted ) = ();

  while ( 6 == ( ( $sNodeName, $sDBServerName, $sInformixDir, $sSDSInstances, $sServerType, $sResGroup) = splice ( @lsEnq, 0, 6 ) ) ) {

    ( $sRes, $sGroup ) = $sResGroup =~ /([^:]+)/g;

    #----------------------------------------------------------------------------
    # Get the Current Primary DB Name from $Informixdir/.VRTSInfxSDPrimDB file
    #----------------------------------------------------------------------------

    $sSDSFile = "$sInformixDir/.VRTSInfxSDPrimDB";
    #----------------------
    # Read in the SDS file..
    #----------------------
    if ( open ( SDSFILE, $sSDSFile ) ) {
      if ( ( $sPrimaryDB ) = map { m/(\S+)/ } <SDSFILE> ) {
        &logTrace(20207,
        "<%s>Got Priamry DB [%s] from SDS file [%s] :",
        $sMyName, $sPrimaryDB, $sSDSFile );
      }
      else {
        &logTrace(20208,
        "<%s>Failed to get any SDS Nodes from SDS file[%s].",
        $sMyName, $sSDSFile );

        $iReturnCode = 0; # Failure..
      }

      #------------------------
      # Close file..
      #------------------------
      unless ( close ( SDSFILE ) ) {
        &logTrace( 20209,
        "<%s> Could not close SDS file [%s] after read: [%s]",
        $sMyName, $sSDSFile, $! );
  
        $iReturnCode = 0; # Failure..
      }
    }
    else {
      &logTrace( 20210,
      "<%s> Could not open SDS file for read:",
      $sMyName, $sSDSFile, $! );
  
      $iReturnCode = 0; # Failure..
    }

    #------------------------------------------------------------------------
    # Next if Server Type is not SDS
    # OR Service Group is not the one for which Preonline trigger is called.
    #------------------------------------------------------------------------

    if ( $sServerType !~ /SharedDisk/ || $sGroup ne $sGrp ){
      next;
    }

    chomp($sDBServerName);
    if ( $sPrimaryDB !~ /$sDBServerName/ ){
      &logTrace(20211, "<%s> Primary DB [%s] is running. No promotion required.", 
      $sMyName, $sDBServerName );

      last; # Do nothing as Primary DB is still up
    }
    else {
      &logTrace(20212, "<%s> Primary DB [%s] is Down! Need to Promote SD Secondary to Primary.", 
      $sMyName, $sDBServerName );
    }

    my %hsSDSInstances = map { $_ => 1 } $sSDSInstances =~ /[^\s]+/g;

    &logTrace(20213, "<%s>Faulted Node : 
    NodeName [%s], Server Type [%s], Resource Name [%s], Service Group Name [%s], 
    DBServer Name [%s], SDS Instances [%s]",
    $sMyName, $sNodeName, $sServerType, $sRes, $sGroup, 
    $sDBServerName, join(', ',keys %hsSDSInstances) );

    # Delete self from SDS Instance list

    if ( $hsSDSInstances { $sDBServerName } ) { 
      delete $hsSDSInstances { $sDBServerName };
    }

    my $sOldPrimDB = $sDBServerName;
    @lsCpy = @lsHaresOut; # Re-initalize Data

    $i = 0;
    @lsNext =
    grep {
      ( s/\n// || 1 )                           && # Chomp, and always make true..
      ( /^(\S+).*Group.*\s+(\S+)/               && # Mark resource, and group name..
      ( $sGroup = "$1:$2" ) && 0 )              || # Store, and always make false..
      ( s/.*\s(\S+)\s+ONLINE$/$1/ && ++ $i )    || # If online, get node name, set flag..
      ( $i                                      &&
      s/.*\s+DBServerName\s+global\s+(\S+)/$1/  &&
      ( $hsSDSInstances { $1 } ) )              ||
      ( $i                                      && # If flag, use demarker..
      s/#/sprintf ( $sGroup, -- $i )/e )           # to print the group name.. reset flag..
    }@lsCpy;

    while ( 3 == ( ( $sNodeName, $sDBServerName, $sResGroup) = splice ( @lsNext, 0, 3 ) ) ) {
       my ( $sRes, $sGroup ) = $sResGroup =~ /([^:]+)/g;

       &logTrace (20214, "<%s>Node to consider for Switching SD Secondary to SD Primary:
       NodeName [%s], Resource Name [%s], Service Group Name [%s], DBServer Name [%s]",
       $sMyName, $sNodeName, $sRes, $sGroup, $sDBServerName );

      #----------------------
      # Form action command
      #----------------------

      $sCmd = $sCmdHares; # Get global..
      $sCmdArgs = "-action $sRes switchprim -actionargs $sOldPrimDB -sys $sNodeName";
  
      #----------------------------------
      # Fire command and collect output..
      #----------------------------------
      grep { s/\n// || 1 } readpipe ( "$sCmd $sCmdArgs" );
  
      #--------------------
      # Check exit status..
      #--------------------
      if ( $iExitCode = $? >> 8 ) {
        &logTrace ( 20215,
        "<%s> Command [%s] with Args [%s] failed with exitcode [%s].",
        $sMyName, $sCmd, $sCmdArgs, $iExitCode );
      }
      else {
        &logTrace ( 20216,
        "<%s>New Primary DB [%s] is successfully up",
        $sMyName, $sDBServerName );
        last;
      }
    }
  }
  }
  
  #-----------------------------------------------------
  # Failover the Faulted Group on the target system 
  #-----------------------------------------------------
  if ( Informix::OnlineGrp ( $sGrp,
                             $sSys ) ) {
    &logInfo ( 20217, "<%s> Issued online of Group [%s] on node [%s]",
    $sMyName, $sGrp, $sSys );
  }
  else {
    &logError ( 20218,
    "<%s> Could not issue online of Group [%s] on node [%s]",
    $sMyName, $sGrp, $sSys );
    $iReturnCode = 0; # Do nothing..
  }

  return $iReturnCode;
}

=pod

 --------------------------------------------------------------------------
 Function : Integer CreateSDSPrimFile ( IN String InformixDir,
                                        IN String DBServerName,
                                        IN bool IsPriamry );

 Description : Creates file with contains the dbsever name of SD 
               Primary instance.

 Parameters :

 $sInformixDir     Absolute path to installation directory
 $sDBServerName    DBServerName of the instaces with needs to put in file
 $bIsPriamry       1 if current DBServer is SD Priarmy

 Return Value: integer value in the set of {0,n} where
               0 if file creation failed
               1 if file successfully created
 --------------------------------------------------------------------------

=cut

sub CreateSDSPrimFile {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sInformixDir,
       $sDBServerName,
       $bIsPrimary ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20219, "Subroutine <%s> called with:
  InformixDir        [%s]
  DBServerName       [%s]",

  $sMyName, $sInformixDir, $sDBServerName );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sSDSFile, $sCmd, $sCmdArgs, $i, 
       $iExitCode, @liServer ) = ();
  
  #--------------------------------------------
  # Request to create file if Primary Server..
  #--------------------------------------------

  $sSDSFile = "$sInformixDir/.VRTSInfxSDPrimDB";
  if ( $bIsPrimary && ! -e $sSDSFile ) {
    #---------
    # Create..
    #---------
    &logTrace ( 20220,
    "<%s> Creating SDS file [%s].", 
    $sMyName, $sSDSFile );
 
    if ( sysopen ( SDSFILE, $sSDSFile, O_WRONLY|O_SYNC|O_CREAT|O_TRUNC,
                   0600 ) ) {
      print SDSFILE $sDBServerName, "\n"; 
  
      &logTrace( 20221,
      "<%s> SDS file [%s] successfully created.",
      $sMyName, $sSDSFile );

      unless ( close ( SDSFILE ) ) {
        &logError ( 20222,
        "<%s> Failed to close SDS file [%s] after update: [%s]",
        $sMyName, $sSDSFile, $! );
        $iReturnCode = 0;
      }
    }
    else {
      &logError ( 20223,
      "<%s> Failed to create SDS file [%s]: [%s]",
      $sMyName, $sSDSFile, $! );
      $iReturnCode = 0;
    }
  }
  else {
    &logTrace( 20224,
    "<%s> SDS file [%s] already exists",
    $sMyName, $sSDSFile );
  }
  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : bool IsPrimary ( IN String EnvFile,
                             IN String InformixDir,
                             IN String DBServerName );

 Description : Check if the Current DB Instance is Primary Server

 Parameters :

 $sEnvFile         Environment File 
 $sInformixDir     Absolute path to installation directory
 $sDBServerName    DBServerName of the instaces with needs to put in file

 Returns : boolean value in the set of {0,1} where

           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub IsPrimary {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sInformixDir, 
       $sDBServerName ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20225, "Subroutine <%s> called with:
  EnvFile            [%s]
  InformixDir        [%s]
  DBServerName       [%s]",
  $sMyName, $sEnvFile, $sInformixDir, $sDBServerName );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sCmd, $sCmdArgs, $i, $iExitCode, $iServers, 
       @liServer, @lsCmdout, $sOutput, $sState, $sMode,
       $sPrimDBName ) = ();

  my $iTimeout = 100;
  $sCmd = "$sInformixDir/bin/onstat";
  
  #------------ Do not indent ---------
  $sCmdArgs = "-i << EOF
sds verbose
q
EOF";
  #------------ Do not indent ---------

  if ( $iReturnCode = Informix::ExecLocalCmd ( "informix",
                                               $sEnvFile,
                                               $sCmd,
                                               $sCmdArgs,
                                               \$iExitCode,
                                               $iTimeout,
                                               \@lsCmdout ) ) {
    if ( $lsCmdout[0] && $lsCmdout [0] =~
         /WARNING: IBM Informix Dynamic Server is no longer running/ ) {

      &logTrace ( 20226,
      "<%s> Cmd [%s] with args [%s] reported the following output [%s]. Returning [%s]\n",
      $sMyName, $sCmd, $sCmdArgs, $lsCmdout [0], $iReturnCode );

      $iReturnCode = 0; # Failure..
    }
    else {
      #---------------
      # Parse output..
      #---------------

      ( $sOutput ) = grep { /\S/ } @lsCmdout;

      if ( ( $sState, $sMode ) = $sOutput
        =~ m/--\s+(.*)\s+\((.*?)\)/ ) { # Configured..

        &logTrace ( 20227,
        "<%s> Found database in State [%s] acting as [%s].", 
        $sMyName, $sState, $sMode);
      }
      
      ( $sPrimDBName ) = 
      map {
        /^Updater node alias name :(.*)/
      } @lsCmdout ;

      if ( $sPrimDBName && ( $sPrimDBName eq $sDBServerName ) ) {
        ( $iServers ) =
        map {
            /^Number of SDS servers:(.*)/
        } @lsCmdout ;
       
        if ( $iServers gt 0 ){
          ( @liServer ) =
          map {
            /(.*)\sActive\s.*Connected/
          } @lsCmdout ;
     
          &logTrace ( 20228,
          "<%s> Total [%s] SD Secondary are [%s]", 
          $sMyName, $iServers, join(',',@liServer));
        }
      }
      else {
        &logTrace ( 20229,
        "<%s> Not a SD Primary Instance.\n", 
        $sMyName);
        $iReturnCode = 0;
      }
    }
  }
  else { # Failed in execution..
    &logTrace ( 20230,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode );
    $iReturnCode = 0;
  }
  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : bool OnlineGrp ( IN String Grp,
                             IN String Sys );

 Description : Online the Faulted Group at the end of PreOnlineTrigger

 Parameters :
 $Grp                 Cluster group to online
 $sSys                System where the group is asked to be made online

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub OnlineGrp {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sGrp,
       $sSys ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20231, "Subroutine <%s> called with:
  Grp             [%s]
  Sys             [%s]",
  $sMyName, $sGrp, $sSys );

  #-------------------------------------------
  # Set return code optimistically to online..
  #-------------------------------------------
  my $iReturnCode = 1; # Success..

  #-------------------------
  # Define local variables..
  #-------------------------
  my $sCmd = $sCmdHagrp; # Get global..
  my @lsCmdArgs = ( "-online", "-nopre", "$sGrp", "-sys", "$sSys" );
  my $iExitCode = undef;

  &logTrace ( 20232, "<%s> Executing Command [%s] with arguments [%s]",
  $sMyName, $sCmd, join( ' ', @lsCmdArgs ) );

  #-------------------------------
  # Validate and execute command..
  #-------------------------------
  if ( ! -e $sCmd ) {
    &logError ( 20233,
    "<%s> Cmd [%s] does not exist", $sMyName, $sCmd );
    $iReturnCode = 0; # Failure..
  }
  elsif ( ! -x $sCmd ) {
    &logError ( 20234,
    "<%s> Cmd [%s] is not executable", $sMyName, $sCmd );
    $iReturnCode = 0; # Failure..
  }
  else { # Execute..
    &logTrace ( 20235,
    "<%s> Executing Cmd [%s] with args [%s]",
    $sMyName, $sCmd, join ( ' ', @lsCmdArgs ) );

    system ( $sCmd, @lsCmdArgs ); # Execute..

    #-------------------------
    # Check execution status..
    #-------------------------
    if ( $iExitCode = $? >> 8 ) {
      &logError ( 20236,
      "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
      $sMyName, $sCmd, join ( ' ', @lsCmdArgs ), $iExitCode );
      $iReturnCode = 0; # Failure..
    }
    else {
      &logTrace ( 20237,
      "<%s> Cmd [%s] with args [%s] executed successfully with exit code [%s]",
      $sMyName, $sCmd, join ( ' ', @lsCmdArgs ), $iExitCode );
    }
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : bool PromoteSDS ( IN String InformixDir,
                              IN String DBServerName
                              IN String SDSInstances );

 Description : Promote SDS if the current offline resource is primary 
               Informix instance.

 Parameters :

 $sInformixDir     Absolute path to installation directory
 $sDBServerName    DBServerName of the instaces with needs to put in file
 $sSDSInstances    List of connected SDS Instances

 Returns : boolean value in the set of {0,1} where

           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub PromoteSDS {

  #----------------
  # Get Arguments..
  #----------------
  my ( $sInformixDir,
       $sDBServerName,
       $sSDSInstances ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  my @lSDSInstances = @$sSDSInstances;

  &logTrace ( 20253, "Subroutine <%s> called with:
  InformixDir              [%s]
  DBServerName             [%s]
  SDSInstances             [%s]",
  $sMyName, $sInformixDir, $sDBServerName, join(',',@lSDSInstances ));

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  my ( $sSDSFile, $sPrimaryDB, $sCmd, $sCmdArgs, $i ) = ();

  $sSDSFile = "$sInformixDir/.VRTSInfxSDPrimDB";
  #----------------------
  # Read in the SDS file..
  #----------------------

  if ( open ( SDSFILE, $sSDSFile ) ) {
    if ( ( $sPrimaryDB ) = map { m/(\S+)/ } <SDSFILE> ) {
      &logTrace( 20254, 
      "<%s>Got Priamry DB [%s] from SDS file [%s].",
      $sMyName, $sPrimaryDB, $sSDSFile );
    }
    else {
      &logTrace(20255,
      "<%s>Failed to get any SDS Nodes from SDS file[%s].",
      $sMyName, $sSDSFile );

      $iReturnCode = 0; # Failure..
    }
    #------------------------
    # Close file..
    #------------------------
    unless ( close ( SDSFILE ) ) {
      &logTrace( 20256,
      "<%s> Could not close SDS file [%s] after read: [%s]",
      $sMyName, $sSDSFile, $! );

      $iReturnCode = 0; # Failure..
    }
  }
  else {
    &logTrace( 20257,
    "<%s> Could not open SDS file for read:",
    $sMyName, $sSDSFile, $! );

    $iReturnCode = 0; # Failure..
  }

  chomp($sDBServerName);
  if ( $sPrimaryDB ne $sDBServerName ){
    &logTrace ( 20258, 
    "<%s> Primary DB [%s] is running. No promotion required.",
    $sMyName, $sDBServerName );

    return $iReturnCode; # Do nothing as Primary DB is still up
  }
  else {
    &logTrace( 20259, 
    "<%s>Primary DB [%s] is Down! Need to Promote SD Secondary to Primary.",
    $sMyName, $sDBServerName );
  }

  #--------------------------
  # Get the attribute values 
  #--------------------------
  $sCmd = $sCmdHares; # Get global..
  $sCmdArgs = "-display -type Informix -attribute " .
              "State Group DBServerName ServerType SDSInstances";

  #---------------------------------------------
  # Assume the group name is valid and proceed..
  #---------------------------------------------
  my ( @lsHaresOut, %hsSDSInstances, $sOldPrimDB, @lsCpy, @lsNext,
       $sResGroup, $sRes, $sGroup, $sNodeName, $iExitCode ) = ();

  @lsHaresOut = readpipe ( "$sCmd $sCmdArgs" );
  push @lsHaresOut, "#"; # Insert trailing de-marker..

  %hsSDSInstances = map { $_ => 1 }@lSDSInstances; 
    
  $sOldPrimDB = $sDBServerName;
  @lsCpy = @lsHaresOut; # Take backup

  $i = 0;
  @lsNext =
  grep {
    ( s/\n// || 1 )                           && # Chomp, and always make true..
    ( /^(\S+).*Group.*\s+(\S+)/               && # Mark resource, and group name..
    ( $sGroup = "$1:$2" ) && 0 )              || # Store, and always make false..
    ( s/.*\s(\S+)\s+ONLINE$/$1/ && ++ $i )    || # If online, get node name, set flag..
    ( $i                                      &&
    s/.*\s+DBServerName\s+global\s+(\S+)/$1/  &&
    ( $hsSDSInstances { $1 } ) )              ||
    ( $i                                      && # If flag, use demarker..
    s/#/sprintf ( $sGroup, -- $i )/e )           # to print the group name.. reset flag..
  }@lsCpy;

  while ( 3 == ( ( $sNodeName, $sDBServerName, $sResGroup) = splice ( @lsNext, 0, 3 ) ) ) {
    ( $sRes, $sGroup ) = $sResGroup =~ /([^:]+)/g;

    if ( $sOldPrimDB eq $sDBServerName ) {  # Skip Self
      next;
    }

    &logTrace ( 20260, 
    "<%s>Node to consider for Switching SD Secondary to SD Primary:
    NodeName [%s], Resource Name [%s], Service Group Name [%s], DBServer Name [%s]",
    $sMyName, $sNodeName, $sRes, $sGroup, $sDBServerName );

    #----------------------
    # Form action command
    #----------------------

    $sCmdArgs = "-action $sRes switchprim -actionargs $sOldPrimDB -sys $sNodeName";

    #---------------------
    # Fire action command 
    #---------------------
    grep { s/\n// || 1 } readpipe ( "$sCmd $sCmdArgs" );

    #--------------------
    # Check exit status..
    #--------------------
    if ( $iExitCode = $? >> 8 ) {
      &logTrace ( 20261,
      "<%s> Command [%s] with Args [%s] failed with exitcode [%s].",
      $sMyName, $sCmd, $sCmdArgs, $iExitCode );

      $iReturnCode = 0;
    }
    else {
      &logTrace ( 20262, 
      "<%s>Successfully Promoted to Primary DB [%s].",
      $sMyName, $sDBServerName );
      last;
    }
  }
  return $iReturnCode;
}

1; # So PERL knows its included..
